# -*- coding: utf-8 -*-
# ui_gtk.py
#
# Copyright © 2008-2010 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Thanks to Pier Carteri <m3tr0@dei.unipd.it> for program Py_Shell.py
# Thanks to Milad Rastian for program pySQLiteGUI
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import shutil
import sys
import os
from os.path import join, isfile, isabs, splitext
import logging
import traceback

from pyglossary.text_utils import urlToPath
from pyglossary.os_utils import click_website
from pyglossary.glossary import *
from .base import *
from pyglossary import core

import gi
gi.require_version("Gtk", "3.0")

from .gtk3_utils import *
from .gtk3_utils.utils import *
from .gtk3_utils.dialog import MyDialog
from .gtk3_utils.resize_button import ResizeButton

# from gi.repository import GdkPixbuf

log = logging.getLogger("root")

gtk.Window.set_default_icon_from_file(logo)

_ = str  # later replace with translator function

pixDir = join(dataDir, "res")  # FIXME


def getCopressedFileExt(fpath):
	fname, ext = splitext(fpath.lower())
	if ext in (".gz", ".bz2", ".zip"):
		fname, ext = splitext(fname)
	return ext


def buffer_get_text(b):
	return b.get_text(
		b.get_start_iter(),
		b.get_end_iter(),
		True,
	)


def pack(box, child, expand=False, fill=False, padding=0):
	if isinstance(box, gtk.Box):
		box.pack_start(child, expand, fill, padding)
	elif isinstance(box, gtk.CellLayout):
		box.pack_start(child, expand)
	else:
		raise TypeError("pack: unkown type %s" % type(box))


def imageFromFile(path):  # the file must exist
	if not isabs(path):
		path = join(pixDir, path)
	im = gtk.Image()
	try:
		im.set_from_file(path)
	except:
		myRaise()
	return im


def color_parse(colorStr):
	rgba = gdk.RGBA()
	if not rgba.parse(colorStr):
		raise ValueError("bad color string %r" % colorStr)
	return rgba.to_color()


class FormatComboBox(gtk.ComboBox):
	def __init__(self):
		gtk.ComboBox.__init__(self)
		self.model = gtk.ListStore(
			str,  # format name, hidden
			# GdkPixbuf.Pixbuf,# icon
			str,  # format description, shown
		)
		self.set_model(self.model)

		cell = gtk.CellRendererText()
		cell.set_visible(False)
		pack(self, cell)
		self.add_attribute(cell, "text", 0)

		# cell = gtk.CellRendererPixbuf()
		# pack(self, cell, False)
		# self.add_attribute(cell, "pixbuf", 1)

		cell = gtk.CellRendererText()
		pack(self, cell, True)
		self.add_attribute(cell, "text", 1)

	def addFormat(self, _format):
		self.get_model().append((
			_format,
			# icon,
			Glossary.formatsDesc[_format],
		))

	def getActive(self):
		index = gtk.ComboBox.get_active(self)
		if index is None or index < 0:
			return ""
		return self.get_model()[index][0]

	def setActive(self, _format):
		for i, row in enumerate(self.get_model()):
			if row[0] == _format:
				gtk.ComboBox.set_active(self, i)
				return


class InputFormatComboBox(FormatComboBox):
	def __init__(self):
		FormatComboBox.__init__(self)
		for _format in Glossary.readFormats:
			self.addFormat(_format)


class OutputFormatComboBox(FormatComboBox):
	def __init__(self):
		FormatComboBox.__init__(self)
		for _format in Glossary.writeFormats:
			self.addFormat(_format)


class GtkTextviewLogHandler(logging.Handler):
	def __init__(self, treeview_dict):
		logging.Handler.__init__(self)

		self.buffers = {}
		for levelname in (
			"CRITICAL",
			"ERROR",
			"WARNING",
			"INFO",
			"DEBUG",
		):
			textview = treeview_dict[levelname]

			buff = textview.get_buffer()
			tag = gtk.TextTag.new(levelname)
			buff.get_tag_table().add(tag)

			self.buffers[levelname] = buff

	def getTag(self, levelname):
		return self.buffers[levelname].get_tag_table().lookup(levelname)

	def setColor(self, levelname, color):  # FIXME
		self.getTag(levelname).set_property("foreground-gdk", color)

	def emit(self, record):
		msg = record.getMessage()
		# msg = msg.replace("\x00", "")

		if record.exc_info:
			_type, value, tback = record.exc_info
			tback_text = "".join(
				traceback.format_exception(_type, value, tback)
			)
			if msg:
				msg += "\n"
			msg += tback_text

		buff = self.buffers[record.levelname]

		buff.insert_with_tags_by_name(
			buff.get_end_iter(),
			msg + "\n",
			record.levelname,
		)


class GtkSingleTextviewLogHandler(GtkTextviewLogHandler):
	def __init__(self, textview):
		GtkTextviewLogHandler.__init__(self, {
			"CRITICAL": textview,
			"ERROR": textview,
			"WARNING": textview,
			"INFO": textview,
			"DEBUG": textview,
		})


class BrowseButton(gtk.Button):
	def __init__(
		self,
		setFilePathFunc,
		label="Browse",
		actionSave=False,
		title="Select File",
	):
		gtk.Button.__init__(self)

		self.set_label(label)
		self.set_image(gtk.Image.new_from_icon_name(
			"document-save" if actionSave else "document-open",
			gtk.IconSize.BUTTON,
		))

		self.actionSave = actionSave
		self.setFilePathFunc = setFilePathFunc
		self.title = title

		self.connect("clicked", self.onClick)

	def onClick(self, widget):
		fcd = gtk.FileChooserDialog(
			transient_for=self.get_toplevel(),
			action=gtk.FileChooserAction.SAVE if self.actionSave
			else gtk.FileChooserAction.OPEN,
			title=self.title,
		)
		fcd.add_button(gtk.STOCK_CANCEL, gtk.ResponseType.CANCEL)
		fcd.add_button(gtk.STOCK_OK, gtk.ResponseType.OK)
		fcd.connect("response", lambda w, e: fcd.hide())
		fcd.connect(
			"file-activated",
			lambda w: fcd.response(gtk.ResponseType.OK)
		)
		if fcd.run() == gtk.ResponseType.OK:
			self.setFilePathFunc(fcd.get_filename())
		fcd.destroy()


class UI(gtk.Dialog, MyDialog, UIBase):
	def write(self, tag):  # FIXME
		pass

	def status(self, msg):
		# try:
		#	_id = self.statusMsgDict[msg]
		# except KeyError:
		#	_id = self.statusMsgDict[msg] = self.statusNewId
		#	self.statusNewId += 1
		_id = self.statusBar.get_context_id(msg)
		self.statusBar.push(_id, msg)

	def __init__(self, **options):
		gtk.Dialog.__init__(self)
		self.set_title("PyGlossary (Gtk3)")
		self.resize(800, 800)
		self.connect("delete-event", self.onDeleteEvent)
		self.prefPages = []
		# self.statusNewId = 0
		# self.statusMsgDict = {}## message -> id
		#####
		self.pref = {}
		self.pref_load(**options)
		#####
		self.assert_quit = False
		self.path = ""
		self.glos = Glossary(ui=self)
		# ____________________ Tab 1 - Convert ____________________ #
		sizeGroup = gtk.SizeGroup(mode=gtk.SizeGroupMode.HORIZONTAL)
		####
		vbox = gtk.VBox()
		vbox.label = _("Convert")
		vbox.icon = ""  # "*.png"
		self.prefPages.append(vbox)
		######
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Input Format")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.convertInputFormatCombo = InputFormatComboBox()
		pack(hbox, self.convertInputFormatCombo)
		pack(vbox, hbox)
		###
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Input File")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.convertInputEntry = gtk.Entry()
		pack(hbox, self.convertInputEntry, 1, 1)
		button = BrowseButton(
			self.convertInputEntry.set_text,
			label="Browse",
			actionSave=False,
			title="Select Input File",
		)
		pack(hbox, button)
		pack(vbox, hbox)
		##
		self.convertInputEntry.connect(
			"changed",
			self.convertInputEntryChanged,
		)
		#####
		vbox.sep1 = gtk.Label(label="")
		vbox.sep1.show()
		pack(vbox, vbox.sep1)
		#####
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Output Format")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.convertOutputFormatCombo = OutputFormatComboBox()
		pack(hbox, self.convertOutputFormatCombo)
		pack(vbox, hbox)
		###
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Output File")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.convertOutputEntry = gtk.Entry()
		pack(hbox, self.convertOutputEntry, 1, 1)
		button = BrowseButton(
			self.convertOutputEntry.set_text,
			label="Browse",
			actionSave=True,
			title="Select Output File",
		)
		pack(hbox, button)
		pack(vbox, hbox)
		##
		self.convertOutputEntry.connect(
			"changed",
			self.convertOutputEntryChanged,
		)
		#####
		hbox = gtk.HBox(spacing=10)
		label = gtk.Label(label="")
		pack(hbox, label, 1, 1, 10)
		self.convertButton = gtk.Button()
		self.convertButton.set_label("Convert")
		self.convertButton.connect("clicked", self.convertClicked)
		pack(hbox, self.convertButton, 1, 1, 10)
		pack(vbox, hbox, 0, 0, 15)
		# ____________________ Tab 2 - Reverse ____________________ #
		self.reverseStatus = ""
		####
		sizeGroup = gtk.SizeGroup(mode=gtk.SizeGroupMode.HORIZONTAL)
		####
		vbox = gtk.VBox()
		vbox.label = _("Reverse")
		vbox.icon = ""  # "*.png"
		# self.prefPages.append(vbox)
		######
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Input Format")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.reverseInputFormatCombo = InputFormatComboBox()
		pack(hbox, self.reverseInputFormatCombo)
		pack(vbox, hbox)
		###
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Input File")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.reverseInputEntry = gtk.Entry()
		pack(hbox, self.reverseInputEntry, 1, 1)
		button = BrowseButton(
			self.reverseInputEntry.set_text,
			label="Browse",
			actionSave=False,
			title="Select Input File",
		)
		pack(hbox, button)
		pack(vbox, hbox)
		##
		self.reverseInputEntry.connect(
			"changed",
			self.reverseInputEntryChanged,
		)
		#####
		vbox.sep1 = gtk.Label(label="")
		vbox.sep1.show()
		pack(vbox, vbox.sep1)
		#####
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Output Tabfile")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.reverseOutputEntry = gtk.Entry()
		pack(hbox, self.reverseOutputEntry, 1, 1)
		button = BrowseButton(
			self.reverseOutputEntry.set_text,
			label="Browse",
			actionSave=True,
			title="Select Output File",
		)
		pack(hbox, button)
		pack(vbox, hbox)
		##
		self.reverseOutputEntry.connect(
			"changed",
			self.reverseOutputEntryChanged,
		)
		#####
		hbox = gtk.HBox(spacing=3)
		label = gtk.Label(label="")
		pack(hbox, label, 1, 1, 5)
		###
		self.reverseStartButton = gtk.Button()
		self.reverseStartButton.set_label(_("Start"))
		self.reverseStartButton.connect("clicked", self.reverseStartClicked)
		pack(hbox, self.reverseStartButton, 1, 1, 2)
		###
		self.reversePauseButton = gtk.Button()
		self.reversePauseButton.set_label(_("Pause"))
		self.reversePauseButton.set_sensitive(False)
		self.reversePauseButton.connect("clicked", self.reversePauseClicked)
		pack(hbox, self.reversePauseButton, 1, 1, 2)
		###
		self.reverseResumeButton = gtk.Button()
		self.reverseResumeButton.set_label(_("Resume"))
		self.reverseResumeButton.set_sensitive(False)
		self.reverseResumeButton.connect("clicked", self.reverseResumeClicked)
		pack(hbox, self.reverseResumeButton, 1, 1, 2)
		###
		self.reverseStopButton = gtk.Button()
		self.reverseStopButton.set_label(_("Stop"))
		self.reverseStopButton.set_sensitive(False)
		self.reverseStopButton.connect("clicked", self.reverseStopClicked)
		pack(hbox, self.reverseStopButton, 1, 1, 2)
		###
		pack(vbox, hbox, 0, 0, 5)
		#####
		# ____________________________________________________________ #
		notebook = gtk.Notebook()
		self.notebook = notebook
		#########
		for vbox in self.prefPages:
			l = gtk.Label(label=vbox.label)
			l.set_use_underline(True)
			vb = gtk.VBox(spacing=3)
			if vbox.icon:
				vbox.image = imageFromFile(vbox.icon)
				pack(vb, vbox.image)
			pack(vb, l)
			vb.show_all()
			notebook.append_page(vbox, vb)
			try:
				notebook.set_tab_reorderable(vbox, True)
			except AttributeError:
				pass
		#######################
		# notebook.set_property("homogeneous", True)  # not in gtk3 FIXME
		# notebook.set_property("tab-border", 5)  # not in gtk3 FIXME
		# notebook.set_property("tab-hborder", 15)  # not in gtk3 FIXME
		pack(self.vbox, notebook, 0, 0)
		# for i in ui.prefPagesOrder:
		#	try:
		#		j = prefPagesOrder[i]
		#	except IndexError:
		#		continue
		#	notebook.reorder_child(self.prefPages[i], j)
		# ____________________________________________________________ #
		self.consoleTextview = textview = gtk.TextView()
		swin = gtk.ScrolledWindow()
		swin.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
		swin.set_border_width(0)
		swin.add(textview)
		pack(self.vbox, swin, 1, 1)
		###
		handler = GtkSingleTextviewLogHandler(textview)
		log.addHandler(handler)
		###
		textview.override_background_color(
			gtk.StateFlags.NORMAL,
			gdk.RGBA(0, 0, 0, 1),
		)
		###
		handler.setColor("CRITICAL", color_parse("red"))
		handler.setColor("ERROR", color_parse("red"))
		handler.setColor("WARNING", color_parse("yellow"))
		handler.setColor("INFO", color_parse("white"))
		handler.setColor("DEBUG", color_parse("white"))
		###
		textview.get_buffer().set_text("Output & Error Console:\n")
		textview.set_editable(False)
		# ____________________________________________________________ #
		self.progressTitle = ""
		self.progressBar = pbar = gtk.ProgressBar()
		pbar.set_fraction(0)
		# pbar.set_text(_("Progress Bar"))
		# pbar.get_style_context()
		# pbar.set_property("height-request", 20)
		pack(self.vbox, pbar, 0, 0)
		############
		hbox = gtk.HBox(spacing=5)
		clearButton = gtk.Button(
			use_stock=gtk.STOCK_CLEAR,
			always_show_image=True,
			label=_("Clear"),
		)
		clearButton.show_all()
		# image = gtk.Image()
		# image.set_from_stock(gtk.STOCK_CLEAR, gtk.IconSize.MENU)
		# clearButton.add(image)
		clearButton.set_border_width(0)
		clearButton.connect("clicked", self.consoleClearButtonClicked)
		set_tooltip(clearButton, "Clear Console")
		pack(hbox, clearButton, 0, 0)
		####
		# hbox.sepLabel1 = gtk.Label(label="")
		# pack(hbox, hbox.sepLabel1, 1, 1)
		######
		hbox.verbosityLabel = gtk.Label(label=_("Verbosity")+":")
		pack(hbox, hbox.verbosityLabel, 0, 0)
		##
		self.verbosityCombo = combo = gtk.ComboBoxText()
		for level, levelName in enumerate(log.levelNamesCap):
			combo.append_text("%s - %s" % (
				level,
				_(levelName)
			))
		combo.set_active(log.getVerbosity())
		combo.set_border_width(0)
		combo.connect("changed", self.verbosityComboChanged)
		pack(hbox, combo, 0, 0)
		####
		# hbox.sepLabel2 = gtk.Label(label="")
		# pack(hbox, hbox.sepLabel2, 1, 1)
		####
		self.statusBar = sbar = gtk.Statusbar()
		pack(hbox, self.statusBar, 1, 1)
		####
		hbox.resizeButton = ResizeButton(self)
		pack(hbox, hbox.resizeButton, 0, 0)
		######
		pack(self.vbox, hbox, 0, 0)
		# ____________________________________________________________ #
		self.vbox.show_all()
		########
		self.status("Select input file")

	def run(self, editPath=None, readOptions=None):
		if readOptions is None:
			readOptions = {}
		# if editPath:
		#	self.notebook.set_current_page(3)
		#	log.info("Opening file "%s" for edit. please wait..."%editPath)
		#	while gtk.events_pending():
		#		gtk.main_iteration_do(False)
		#	self.dbe_open(editPath, **readOptions)
		gtk.Dialog.present(self)
		gtk.main()

	def onDeleteEvent(self, widget, event):
		self.destroy()
		gtk.main_quit()

	def consoleClearButtonClicked(self, widget=None):
		self.consoleTextview.get_buffer().set_text("")

	def verbosityComboChanged(self, widget=None):
		verbosity = self.verbosityCombo.get_active()
		# or int(self.verbosityCombo.get_active_text())
		log.setVerbosity(verbosity)

	def convertClicked(self, widget=None):
		inPath = self.convertInputEntry.get_text()
		if not inPath:
			self.status("Input file path is empty!")
			log.critical("Input file path is empty!")
			return
		inFormat = self.convertInputFormatCombo.getActive()
		if inFormat:
			inFormatDesc = Glossary.formatsDesc[inFormat]
		else:
			inFormatDesc = ""
			# log.critical("Input format is empty!");return

		outPath = self.convertOutputEntry.get_text()
		if not outPath:
			self.status("Output file path is empty!")
			log.critical("Output file path is empty!")
			return
		outFormat = self.convertOutputFormatCombo.getActive()
		if outFormat:
			outFormatDesc = Glossary.formatsDesc[outFormat]
		else:
			outFormatDesc = ""
			# log.critical("Output format is empty!");return

		while gtk.events_pending():
			gtk.main_iteration_do(False)

		self.convertButton.set_sensitive(False)
		self.progressTitle = "Converting"
		try:
			# if inFormat=="Omnidic":
			#	dicIndex = self.xml.get_widget("spinbutton_omnidic_i")\
			#		.get_value_as_int()
			#	ex = self.glos.readOmnidic(inPath, dicIndex=dicIndex)
			# else:
			finalOutputFile = self.glos.convert(
				inPath,
				inputFormat=inFormat,
				outputFilename=outPath,
				outputFormat=outFormat,
			)
			if finalOutputFile:
				self.status("Convert finished")
			else:
				self.status("Convert failed")
			return bool(finalOutputFile)

		finally:
			self.convertButton.set_sensitive(True)
			self.assert_quit = False
			self.progressTitle = ""

		return True

	def convertInputEntryChanged(self, widget=None):
		inPath = self.convertInputEntry.get_text()
		inFormat = self.convertInputFormatCombo.getActive()
		if inPath.startswith("file://"):
			inPath = urlToPath(inPath)
			self.convertInputEntry.set_text(inPath)

		inExt = getCopressedFileExt(inPath)
		inFormatNew = Glossary.extFormat.get(inExt)

		if not inFormatNew:
			return

		if not isfile(inPath):
			return

		if self.pref["ui_autoSetFormat"] and not inFormat:
			self.convertInputFormatCombo.setActive(inFormatNew)

		self.status("Select output file")
		if self.pref["ui_autoSetOutputFileName"]:
			outFormat = self.convertOutputFormatCombo.getActive()
			outPath = self.convertOutputEntry.get_text()
			if outFormat:
				if not outPath and "." in inPath:
					outPath = splitext(inPath)[0] + \
						Glossary.formatsExt[outFormat][0]
					self.convertOutputEntry.set_text(outPath)
					self.status("Press \"Convert\"")

	def convertOutputEntryChanged(self, widget=None):
		outPath = self.convertOutputEntry.get_text()
		outFormat = self.convertOutputFormatCombo.getActive()
		if not outPath:
			return
		# outFormat = self.combobox_o.get_active_text()
		if outPath.startswith("file://"):
			outPath = urlToPath(outPath)
			self.convertOutputEntry.set_text(outPath)

		if self.pref["ui_autoSetFormat"] and not outFormat:
			outExt = getCopressedFileExt(outPath)
			try:
				outFormatNew = Glossary.extFormat[outExt]
			except KeyError:
				pass
			else:
				self.convertOutputFormatCombo.setActive(outFormatNew)

		if self.convertOutputFormatCombo.getActive():
			self.status("Press \"Convert\"")
		else:
			self.status("Select output format")

	def reverseLoad(self):
		pass

	def reverseStartLoop(self):
		pass

	def reverseStart(self):
		if not self.reverseLoad():
			return
		###
		self.reverseStatus = "doing"
		self.reverseStartLoop()
		###
		self.reverseStartButton.set_sensitive(False)
		self.reversePauseButton.set_sensitive(True)
		self.reverseResumeButton.set_sensitive(False)
		self.reverseStopButton.set_sensitive(True)

	def reverseStartClicked(self, widget=None):
		self.waitingDo(self.reverseStart)

	def reversePause(self):
		self.reverseStatus = "pause"
		###
		self.reverseStartButton.set_sensitive(False)
		self.reversePauseButton.set_sensitive(False)
		self.reverseResumeButton.set_sensitive(True)
		self.reverseStopButton.set_sensitive(True)

	def reversePauseClicked(self, widget=None):
		self.waitingDo(self.reversePause)

	def reverseResume(self):
		self.reverseStatus = "doing"
		###
		self.reverseStartButton.set_sensitive(False)
		self.reversePauseButton.set_sensitive(True)
		self.reverseResumeButton.set_sensitive(False)
		self.reverseStopButton.set_sensitive(True)

	def reverseResumeClicked(self, widget=None):
		self.waitingDo(self.reverseResume)

	def reverseStop(self):
		self.reverseStatus = "stop"
		###
		self.reverseStartButton.set_sensitive(True)
		self.reversePauseButton.set_sensitive(False)
		self.reverseResumeButton.set_sensitive(False)
		self.reverseStopButton.set_sensitive(False)

	def reverseStopClicked(self, widget=None):
		self.waitingDo(self.reverseStop)

	def reverseInputEntryChanged(self, widget=None):
		inPath = self.reverseInputEntry.get_text()
		inFormat = self.reverseInputFormatCombo.getActive()
		if inPath.startswith("file://"):
			inPath = urlToPath(inPath)
			self.reverseInputEntry.set_text(inPath)

		inExt = getCopressedFileExt(inPath)
		inFormatNew = Glossary.extFormat.get(inExt)

		if inFormatNew and self.pref["ui_autoSetFormat"] and not inFormat:
			self.reverseInputFormatCombo.setActive(inFormatNew)

		if self.pref["ui_autoSetOutputFileName"]:
			outExt = ".txt"
			outPath = self.reverseOutputEntry.get_text()
			if inFormatNew and not outPath:
				outPath = splitext(inPath)[0] + "-reversed" + outExt
				self.reverseOutputEntry.set_text(outPath)

	def reverseOutputEntryChanged(self, widget=None):
		pass

	def progressInit(self, title):
		self.progressTitle = title

	def progress(self, rat, text=None):
		if not text:
			text = "%%%d" % (rat*100)
		text += " - %s" % self.progressTitle
		self.progressBar.set_fraction(rat)
		# self.progressBar.set_text(text)  # not working
		self.status(text)
		while gtk.events_pending():
			gtk.main_iteration_do(False)



# -*- coding: utf-8 -*-

import re

from .text_utils import (
	fixUtf8,
)


class EntryFilter(object):
	name = ''
	desc = ''

	def __init__(self, glos):
		self.glos = glos

	def run(self, entry):
		"""
			returns an Entry object, or None to skip
				may return the same `entry`,
				or modify and return it,
				or return a new Entry object
		"""
		return entry


class StripEntryFilter(EntryFilter):
	name = 'strip'
	desc = 'Strip Whitespaces'

	def run(self, entry):
		entry.strip()
		entry.replace('\r', '')
		return entry


class NonEmptyWordFilter(EntryFilter):
	name = 'non_empty_word'
	desc = 'Non-empty Words'

	def run(self, entry):
		if not entry.getWord():
			return
#		words = entry.getWords()
#		if not words:
#			return
#		wordsStr = ''.join([w.strip() for w in words])
#		if not wordsStr:
#			return
		return entry


class NonEmptyDefiFilter(EntryFilter):
	name = 'non_empty_defi'
	desc = 'Non-empty Definition'

	def run(self, entry):
		if not entry.getDefi():
			return
		return entry


class FixUnicodeFilter(EntryFilter):
	name = 'fix_unicode'
	desc = 'Fix Unicode'

	def run(self, entry):
		entry.editFuncWord(fixUtf8)
		entry.editFuncDefi(fixUtf8)
		return entry


class LowerWordFilter(EntryFilter):
	name = 'lower_word'
	desc = 'Lowercase Words'

	def run(self, entry):
		entry.editFuncWord(str.lower)
		return entry


class SkipDataEntryFilter(EntryFilter):
	name = 'skip_resources'
	desc = 'Skip Resources'

	def run(self, entry):
		if entry.isData():
			return
		return entry


class LangEntryFilter(EntryFilter):
	name = 'lang'
	desc = 'Language-dependent Filters'

	def run_fa(self, entry):
		from pyglossary.persian_utils import faEditStr
		entry.editFuncWord(faEditStr)
		entry.editFuncDefi(faEditStr)
		# RLM = '\xe2\x80\x8f'
		# defi = '\n'.join([RLM+line for line in defi.split('\n')])
		# for GoldenDict ^^ FIXME
		return entry

	def run(self, entry):
		langs = (
			self.glos.getInfo('sourceLang') +
			self.glos.getInfo('targetLang')
		).lower()
		if 'persian' in langs or 'farsi' in langs:
			entry = self.run_fa(entry)

		return entry


class CleanEntryFilter(EntryFilter):  # FIXME
	name = 'clean'
	desc = 'Clean'

	def cleanDefi(self, st):
		st = st.replace('♦  ', '♦ ')
		st = re.sub('[\r\n]+', '\n', st)
		st = re.sub(' *\n *', '\n', st)

		"""
		This code may correct snippets like:
		- First sentence .Second sentence. -> First sentence. Second sentence.
		- First clause ,second clause. -> First clause, second clause.
		But there are cases when this code have undesirable effects
		( '<' represented as '&lt;' in HTML markup):
		- <Adj.> -> < Adj. >
		- <fig.> -> < fig. >
		"""
		"""
		for j in range(3):
			for ch in ',.;':
				st = replacePostSpaceChar(st, ch)
		"""

		st = re.sub('♦\n+♦', '♦', st)
		if st.endswith('<p'):
			st = st[:-2]
		st = st.strip()
		if st.endswith(','):
			st = st[:-1]

		return st

	def run(self, entry):
		entry.editFuncDefi(self.cleanDefi)
		return entry



from formats_common import *

import sys
import os
from os.path import (
	join,
	split,
	splitext,
	isfile,
	isdir,
	exists,
)

import logging

log = logging.getLogger("root")

from pprint import pformat

from paths import rootDir
sys.path.insert(0, rootDir)

from pyglossary.flags import *

from pyglossary import core
from pyglossary.file_utils import FileLineWrapper
from pyglossary.text_utils import toStr, toBytes
from pyglossary.os_utils import indir

enable = False
format = "Unknown"
description = "Unknown"
extentions = []
readOptions = []
writeOptions = []
supportsAlternates = False
sortOnWrite = DEFAULT_NO
sortKey = None



from .text_utils import replacePostSpaceChar


def faEditStr(st):
	return replacePostSpaceChar(
		st.replace('ي', 'ی')
		  .replace('ك', 'ک')
		  .replace('ۂ', 'هٔ')
		  .replace('ہ', 'ه'),
		'،',
	)



# -*- coding: utf-8 -*-

# Source Glossary for "Babylon Builder".
# A plain text file. Not binary like BGL files.

from formats_common import *

enable = True
format = "BabylonSource"
description = "Babylon Source (gls)"
extentions = [".gls", ".babylon"]
readOptions = []
writeOptions = [
	"writeInfo",  # bool
	"newline",  # str, or choice ("\r\n", "\n", or "\r")
	"encoding",  # str
	"resources",  # bool
]


def entryCleanWinArabic(entry):
	from pyglossary.arabic_utils import cleanWinArabicStr
	entry.editFuncWord(cleanWinArabicStr)
	entry.editFuncDefi(cleanWinArabicStr)
	return entry


def write(
	glos,
	filename,
	writeInfo=True,
	newline="",
	encoding="",
	resources=True,
):
	g = glos
	entryFilterFunc = None
	if encoding.lower() in ("", "utf8", "utf-8"):
		encoding = "UTF-8"
	elif encoding.lower() in (
		"arabic",
		"windows-1256",
		"windows-arabic",
		"arabic-windows",
		"arabic windows",
		"windows arabic",
	):
		encoding = "windows-1256"
		entryFilterFunc = entryCleanWinArabic
		if not newline:
			newline = "\r\n"

	if not newline:
		newline = "\n"

	head = ""
	if writeInfo:
		head += "\n".join([
			"### Glossary title:%s" % g.getInfo("name"),
			"### Author:%s" % g.getInfo("author"),
			"### Description:%s" % g.getInfo("description"),
			"### Source language:%s" % g.getInfo("inputlang"),
			"### Source alphabet:%s" % encoding,
			"### Target language:%s" % g.getInfo("outputlang"),
			"### Target alphabet:%s" % encoding,
			"### Browsing enabled?Yes",
			"### Type of glossary:00000000",
			"### Case sensitive words?0"
			"",
			"### Glossary section:",
			"",
		])

	g.writeTxt(
		"\n",
		"\n\n",
		filename=filename,
		writeInfo=False,
		rplList=(
			("\n", "<BR>"),
		),
		ext=".gls",
		head=head,
		entryFilterFunc=entryFilterFunc,
		encoding=encoding,
		newline=newline,
		resources=resources,
	)



# -*- coding: utf-8 -*-
# copied from https://github.com/mitsuhiko/jinja2-htmlcompress
"""
    jinja2htmlcompress
    ~~~~~~~~~~~~~~~~~~

    A Jinja2 extension that eliminates useless whitespace at template
    compilation time without extra overhead.

    :copyright: (c) 2011 by Armin Ronacher.
    :license: BSD, see bottom of file for more details.
"""
import re
from jinja2.ext import Extension
from jinja2.lexer import Token, describe_token
from jinja2 import TemplateSyntaxError


_tag_re = re.compile(r'(?:<(/?)([a-zA-Z0-9_-]+)\s*|(>\s*))(?s)')
_ws_normalize_re = re.compile(r'[ \t\r\n]+')


class StreamProcessContext(object):

    def __init__(self, stream):
        self.stream = stream
        self.token = None
        self.stack = []

    def fail(self, message):
        raise TemplateSyntaxError(message, self.token.lineno,
                                  self.stream.name, self.stream.filename)


def _make_dict_from_listing(listing):
    rv = {}
    for keys, value in listing:
        for key in keys:
            rv[key] = value
    return rv


class HTMLCompress(Extension):
    isolated_elements = set(['script', 'style', 'noscript', 'textarea'])
    void_elements = set(['br', 'img', 'area', 'hr', 'param', 'input',
                         'embed', 'col'])
    block_elements = set(['div', 'p', 'form', 'ul', 'ol', 'li', 'table', 'tr',
                          'tbody', 'thead', 'tfoot', 'tr', 'td', 'th', 'dl',
                          'dt', 'dd', 'blockquote', 'h1', 'h2', 'h3', 'h4',
                          'h5', 'h6', 'pre'])
    breaking_rules = _make_dict_from_listing([
        (['p'], set(['#block'])),
        (['li'], set(['li'])),
        (['td', 'th'], set(['td', 'th', 'tr', 'tbody', 'thead', 'tfoot'])),
        (['tr'], set(['tr', 'tbody', 'thead', 'tfoot'])),
        (['thead', 'tbody', 'tfoot'], set(['thead', 'tbody', 'tfoot'])),
        (['dd', 'dt'], set(['dl', 'dt', 'dd']))
    ])

    def is_isolated(self, stack):
        for tag in reversed(stack):
            if tag in self.isolated_elements:
                return True
        return False

    def is_breaking(self, tag, other_tag):
        breaking = self.breaking_rules.get(other_tag)
        return breaking and (tag in breaking or
            ('#block' in breaking and tag in self.block_elements))

    def enter_tag(self, tag, ctx):
        while ctx.stack and self.is_breaking(tag, ctx.stack[-1]):
            self.leave_tag(ctx.stack[-1], ctx)
        if tag not in self.void_elements:
            ctx.stack.append(tag)

    def leave_tag(self, tag, ctx):
        if not ctx.stack:
            ctx.fail('Tried to leave "%s" but something closed '
                     'it already' % tag)
        if tag == ctx.stack[-1]:
            ctx.stack.pop()
            return
        for idx, other_tag in enumerate(reversed(ctx.stack)):
            if other_tag == tag:
                for num in range(idx + 1):
                    ctx.stack.pop()
            elif not self.breaking_rules.get(other_tag):
                break

    def normalize(self, ctx):
        pos = 0
        buffer = []
        def write_data(value):
            if not self.is_isolated(ctx.stack):
                value = _ws_normalize_re.sub(' ', value.strip())
            buffer.append(value)

        for match in _tag_re.finditer(ctx.token.value):
            closes, tag, sole = match.groups()
            preamble = ctx.token.value[pos:match.start()]
            write_data(preamble)
            if sole:
                write_data(sole)
            else:
                buffer.append(match.group())
                (closes and self.leave_tag or self.enter_tag)(tag, ctx)
            pos = match.end()

        write_data(ctx.token.value[pos:])
        return ''.join(buffer)

    def filter_stream(self, stream):
        ctx = StreamProcessContext(stream)
        for token in stream:
            if token.type != 'data':
                yield token
                continue
            ctx.token = token
            value = self.normalize(ctx)
            yield Token(token.lineno, 'data', value)


class SelectiveHTMLCompress(HTMLCompress):

    def filter_stream(self, stream):
        ctx = StreamProcessContext(stream)
        strip_depth = 0
        while 1:
            if stream.current.type == 'block_begin':
                if stream.look().test('name:strip') or \
                   stream.look().test('name:endstrip'):
                    stream.skip()
                    if stream.current.value == 'strip':
                        strip_depth += 1
                    else:
                        strip_depth -= 1
                        if strip_depth < 0:
                            ctx.fail('Unexpected tag endstrip')
                    stream.skip()
                    if stream.current.type != 'block_end':
                        ctx.fail('expected end of block, got %s' %
                                 describe_token(stream.current))
                    stream.skip()
            if strip_depth > 0 and stream.current.type == 'data':
                ctx.token = stream.current
                value = self.normalize(ctx)
                yield Token(stream.current.lineno, 'data', value)
            else:
                yield stream.current
            next(stream)


def test():
    from jinja2 import Environment
    env = Environment(extensions=[HTMLCompress])
    tmpl = env.from_string('''
        <html>
          <head>
            <title>{{ title }}</title>
          </head>
          <script type=text/javascript>
            if (foo < 42) {
              document.write('Foo < Bar');
            }
          </script>
          <body>
            <li><a href="{{ href }}">{{ title }}</a><br>Test   Foo
            <li><a href="{{ href }}">{{ title }}</a><img src=test.png>
          </body>
        </html>
    ''')
    print(tmpl.render(title=42, href='index.html'))

    env = Environment(extensions=[SelectiveHTMLCompress])
    tmpl = env.from_string('''
        Normal   <span>  unchanged </span> stuff
        {% strip %}Stripped <span class=foo  >   test   </span>
        <a href="foo">  test </a> {{ foo }}
        Normal <stuff>   again {{ foo }}  </stuff>
        <p>
          Foo<br>Bar
          Baz
        <p>
          Moep    <span>Test</span>    Moep
        </p>
        {% endstrip %}
    ''')
    print(tmpl.render(foo=42))


if __name__ == '__main__':
    test()

# What follows is the original LICENSE file; there was not actually an AUTHORS
# file in the original repo, so its absence here should not be an issue.

# Copyright (c) 2011 by Armin Ronacher, see AUTHORS for more details.
#
# Some rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     * The names of the contributors may not be used to endorse or
#       promote products derived from this software without specific
#       prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



# -*- coding: utf-8 -*-
# ui_tk.py
#
# Copyright © 2009-2010 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.


from pyglossary.core import homeDir
from pyglossary.glossary import *
from pyglossary.text_utils import urlToPath
from .base import *
from os.path import join
import logging
import traceback

import tkinter as tk
from tkinter import filedialog
from tkinter import tix

log = logging.getLogger("root")

# startBold = "\x1b[1m"  # Start Bold #len=4
# startUnderline = "\x1b[4m"  # Start Underline #len=4
endFormat = "\x1b[0;0;0m"  # End Format #len=8
# redOnGray = "\x1b[0;1;31;47m"
startRed = "\x1b[31m"


noneItem = "Not Selected"


bitmapLogo = join(dataDir, "res", "pyglossary.ico") if "nt" == os.name \
	else "@" + join(dataDir, "res", "pyglossary.xbm")


def set_window_icon(window):
	# window.wm_iconbitmap(bitmap=bitmapLogo)
	window.iconphoto(
		True,
		tk.PhotoImage(file=join(dataDir, "res", "pyglossary.png")),
	)


class TkTextLogHandler(logging.Handler):
	def __init__(self, tktext):
		logging.Handler.__init__(self)
		#####
		tktext.tag_config("CRITICAL", foreground="#ff0000")
		tktext.tag_config("ERROR", foreground="#ff0000")
		tktext.tag_config("WARNING", foreground="#ffff00")
		tktext.tag_config("INFO", foreground="#00ff00")
		tktext.tag_config("DEBUG", foreground="#ffffff")
		###
		self.tktext = tktext

	def emit(self, record):
		msg = record.getMessage()
		###
		if record.exc_info:
			_type, value, tback = record.exc_info
			tback_text = "".join(
				traceback.format_exception(_type, value, tback)
			)
			if msg:
				msg += "\n"
			msg += tback_text
		###
		self.tktext.insert(
			"end",
			msg + "\n",
			record.levelname,
		)


# Monkey-patch Tkinter
# http://stackoverflow.com/questions/5191830/python-exception-logging
def CallWrapper__call__(self, *args):
	"""
		Apply first function SUBST to arguments, than FUNC.
	"""
	if self.subst:
		args = self.subst(*args)
	try:
		return self.func(*args)
	except:
		log.exception("Exception in Tkinter callback:")
tk.CallWrapper.__call__ = CallWrapper__call__


class ProgressBar(tix.Frame):
	"""
	This comes from John Grayson's book "Python and Tkinter programming"
	Edited by Saeed Rasooli
	"""
	def __init__(
		self,
		rootWin=None,
		orientation="horizontal",
		min_=0,
		max_=100,
		width=100,
		height=18,
		appearance="sunken",
		fillColor="blue",
		background="gray",
		labelColor="yellow",
		labelFont="Verdana",
		labelFormat="%d%%",
		value=0,
		bd=2,
	):
		# preserve various values
		self.rootWin = rootWin
		self.orientation = orientation
		self.min = min_
		self.max = max_
		self.width = width
		self.height = height
		self.fillColor = fillColor
		self.labelFont = labelFont
		self.labelColor = labelColor
		self.background = background
		self.labelFormat = labelFormat
		self.value = value
		tix.Frame.__init__(self, rootWin, relief=appearance, bd=bd)
		self.canvas = tix.Canvas(
			self,
			height=height,
			width=width,
			bd=0,
			highlightthickness=0,
			background=background,
		)
		self.scale = self.canvas.create_rectangle(
			0,
			0,
			width,
			height,
			fill=fillColor,
		)
		self.label = self.canvas.create_text(
			width/2,
			height/2,
			text="",
			anchor="c",
			fill=labelColor,
			font=self.labelFont,
		)
		self.update()
		self.bind("<Configure>", self.update)
		self.canvas.pack(side="top", fill="x", expand="no")

	def updateProgress(self, newVal, newMax=None, text=""):
		if newMax:
			self.max = newMax
		self.value = newVal
		self.update(None, text)

	def update(self, event=None, labelText=""):
		# Trim the values to be between min and max
		value = self.value
		if value > self.max:
			value = self.max
		if value < self.min:
			value = self.min
		# Adjust the rectangle
		width = int(self.winfo_width())
		# width = self.width
		ratio = float(value)/self.max
		if self.orientation == "horizontal":
			self.canvas.coords(
				self.scale,
				0,
				0,
				width * ratio,
				self.height,
			)
		else:
			self.canvas.coords(
				self.scale,
				0,
				self.height * (1 - ratio),
				width,
				self.height,
			)
		# Now update the colors
		self.canvas.itemconfig(self.scale, fill=self.fillColor)
		self.canvas.itemconfig(self.label, fill=self.labelColor)
		# And update the label
		if not labelText:
			labelText = self.labelFormat % int(ratio * 100)
		self.canvas.itemconfig(self.label, text=labelText)
		# FIXME:
		# self.canvas.move(self.label, width/2, self.height/2)
		# self.canvas.scale(self.label, 0, 0, float(width)/self.width, 1)
		self.canvas.update_idletasks()


class UI(tix.Frame, UIBase):
	def __init__(self, path="", **options):
		self.glos = Glossary(ui=self)
		self.pref = {}
		self.pref_load(**options)
		#############################################
		rootWin = self.rootWin = tix.Tk()
		tix.Frame.__init__(self, rootWin)
		rootWin.title("PyGlossary (Tkinter)")
		rootWin.resizable(True, False)
		########
		set_window_icon(rootWin)
		########
		self.pack(fill="x")
		# rootWin.bind("<Configure>", self.resized)
		######################
		self.glos = Glossary(ui=self)
		self.pref = {}
		self.pref_load()
		self.pathI = ""
		self.pathO = ""
		self.fcd_dir = join(homeDir, "Desktop")
		######################
		vpaned = tk.PanedWindow(self, orient=tk.VERTICAL)
		notebook = tix.NoteBook(vpaned)
		notebook.add("tab1", label="Convert", underline=0)
		notebook.add("tab2", label="Reverse", underline=0)
		convertFrame = tix.Frame(notebook.tab1)
		######################
		frame = tix.Frame(convertFrame)
		##
		label = tix.Label(frame, text="Read from format")
		label.pack(side="left")
		##
		comboVar = tk.StringVar()
		combo = tk.OptionMenu(frame, comboVar, *Glossary.readDesc)
		# comboVar.set(Glossary.readDesc[0])
		comboVar.set(noneItem)
		combo.pack(side="left")
		self.combobox_i = comboVar
		##
		frame.pack(fill="x")
		###################
		frame = tix.Frame(convertFrame)
		##
		label = tix.Label(frame, text="  Path:")
		label.pack(side="left")
		##
		entry = tix.Entry(frame)
		entry.pack(side="left", fill="x", expand=True)
		entry.bind_all("<KeyPress>", self.entry_changed)
		self.entry_i = entry
		##
		button = tix.Button(
			frame,
			text="Browse",
			command=self.browse_i,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.pack(side="left")
		##
		frame.pack(fill="x")
		######################
		frame = tix.Frame(convertFrame)
		##
		label = tix.Label(frame, text="Write to format    ")
		label.pack(side="left")
		##
		comboVar = tk.StringVar()
		combo = tk.OptionMenu(frame, comboVar, *Glossary.writeDesc)
		# comboVar.set(Glossary.writeDesc[0])
		comboVar.set(noneItem)
		combo.pack(side="left")
		combo.bind("<Configure>", self.combobox_o_changed)
		self.combobox_o = comboVar
		##
		frame.pack(fill="x")
		###################
		frame = tix.Frame(convertFrame)
		##
		label = tix.Label(frame, text="  Path:")
		label.pack(side="left")
		##
		entry = tix.Entry(frame)
		entry.pack(side="left", fill="x", expand=True)
		# entry.bind_all("<KeyPress>", self.entry_changed)
		self.entry_o = entry
		##
		button = tix.Button(
			frame,
			text="Browse",
			command=self.browse_o,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.pack(side="left")
		##
		frame.pack(fill="x")
		#######
		frame = tix.Frame(convertFrame)
		label = tix.Label(frame, text=" "*15)
		label.pack(
			side="left",
			fill="x",
			expand=True,
		)
		button = tix.Button(
			frame,
			text="Convert",
			command=self.convert,
			# bg="#00e000",
			# activebackground="#22f022",
		)
		button.pack(
			side="left",
			fill="x",
			expand=True,
		)
		###
		frame.pack(fill="x")
		######
		convertFrame.pack(fill="x")
		vpaned.add(notebook)
		#################
		console = tix.Text(vpaned, height=15, background="#000000")
		# self.consoleH = 15
		# sbar = Tix.Scrollbar(
		#	vpaned,
		#	orien=Tix.VERTICAL,
		#	command=console.yview
		# )
		# sbar.grid ( row=0, column=1)
		# console["yscrollcommand"] = sbar.set
		# console.grid()
		console.pack(fill="both", expand=True)
		log.addHandler(
			TkTextLogHandler(console),
		)
		console.insert("end", "Console:\n")
		####
		vpaned.add(console)
		vpaned.pack(fill="both", expand=True)
		self.console = console
		##################
		frame2 = tix.Frame(self)
		clearB = tix.Button(
			frame2,
			text="Clear",
			command=self.console_clear,
			# bg="black",
			# fg="#ffff00",
			# activebackground="#333333",
			# activeforeground="#ffff00",
		)
		clearB.pack(side="left")
		####
		label = tix.Label(frame2, text="Verbosity")
		label.pack(side="left")
		##
		comboVar = tk.StringVar()
		combo = tk.OptionMenu(
			frame2,
			comboVar,
			0, 1, 2, 3, 4,
		)
		comboVar.set(log.getVerbosity())
		comboVar.trace("w", self.verbosityChanged)
		combo.pack(side="left")
		self.verbosityCombo = comboVar
		#####
		pbar = ProgressBar(frame2, width=400)
		pbar.pack(side="left", fill="x", expand=True)
		self.pbar = pbar
		frame2.pack(fill="x")
		self.progressTitle = ""
		#############
		# vpaned.grid()
		# bottomFrame.grid()
		# self.grid()
		#####################
		# lbox = Tix.Listbox(convertFrame)
		# lbox.insert(0, "aaaaaaaa", "bbbbbbbbbbbbbbbbbbbb")
		# lbox.pack(fill="x")
		##############
		frame3 = tix.Frame(self)
		aboutB = tix.Button(
			frame3,
			text="About",
			command=self.about_clicked,
			# bg="#e000e0",
			# activebackground="#f030f0",
		)
		aboutB.pack(side="right")
		closeB = tix.Button(
			frame3,
			text="Close",
			command=self.quit,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		closeB.pack(side="right")
		frame3.pack(fill="x")
		# __________________________ Reverse Tab __________________________ #
		revFrame = tix.Frame(notebook.tab2)
		revFrame.pack(fill="x")
		######################
		frame = tix.Frame(revFrame)
		##
		label = tix.Label(frame, text="Read from format")
		label.pack(side="left")
		##
		comboVar = tk.StringVar()
		combo = tk.OptionMenu(frame, comboVar, *Glossary.readDesc)
		# comboVar.set(Glossary.readDesc[0])
		comboVar.set(noneItem)
		combo.pack(side="left")
		self.combobox_r_i = comboVar
		##
		frame.pack(fill="x")
		###################
		frame = tix.Frame(revFrame)
		##
		label = tix.Label(frame, text="  Path:")
		label.pack(side="left")
		##
		entry = tix.Entry(frame)
		entry.pack(side="left", fill="x", expand=True)
		# entry.bind_all("<KeyPress>", self.entry_r_i_changed)
		self.entry_r_i = entry
		##
		button = tix.Button(
			frame,
			text="Browse",
			command=self.r_browse_i,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.pack(side="left")
		##
		button = tix.Button(
			frame,
			text="Load",
			command=self.r_load,
			# bg="#7777ff",
		)
		button.pack(side="left")
		###
		frame.pack(fill="x")
		###################
		frame = tix.Frame(revFrame)
		##
		label = tix.Label(frame, text="Output Tabfile")
		label.pack(side="left")
		###
		entry = tix.Entry(frame)
		entry.pack(side="left", fill="x", expand=True)
		# entry.bind_all("<KeyPress>", self.entry_r_i_changed)
		self.entry_r_o = entry
		##
		button = tix.Button(
			frame,
			text="Browse",
			command=self.r_browse_o,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.pack(side="left")
		##
		frame.pack(fill="x")
		##############################
		if path:
			self.entry_i.insert(0, path)
			self.entry_changed()
			self.load()

	def verbosityChanged(self, index, value, op):
		log.setVerbosity(
			int(self.verbosityCombo.get())
		)

	def about_clicked(self):
		about = tix.Toplevel(width=600)  # bg="#0f0" does not work
		about.title("About PyGlossary")
		about.resizable(False, False)
		set_window_icon(about)
		###
		msg1 = tix.Message(
			about,
			width=350,
			text="PyGlossary %s (Tkinter)" % VERSION,
			font=("DejaVu Sans", 13, "bold"),
		)
		msg1.pack(fill="x", expand=True)
		###
		msg2 = tix.Message(
			about,
			width=350,
			text=aboutText,
			font=("DejaVu Sans", 9, "bold"),
			justify=tix.CENTER,
		)
		msg2.pack(fill="x", expand=True)
		###
		msg3 = tix.Message(
			about,
			width=350,
			text=homePage,
			font=("DejaVu Sans", 8, "bold"),
			fg="#3333ff",
		)
		msg3.pack(fill="x", expand=True)
		###
		msg4 = tix.Message(
			about,
			width=350,
			text="Install PyGTK to have a better interface!",
			font=("DejaVu Sans", 8, "bold"),
			fg="#00aa00",
		)
		msg4.pack(fill="x", expand=True)
		###########
		frame = tix.Frame(about)
		###
		button = tix.Button(
			frame,
			text="Close",
			command=about.destroy,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		button.pack(side="right")
		###
		button = tix.Button(
			frame,
			text="License",
			command=self.about_license_clicked,
			# bg="#00e000",
			# activebackground="#22f022",
		)
		button.pack(side="right")
		###
		button = tix.Button(
			frame,
			text="Credits",
			command=self.about_credits_clicked,
			# bg="#0000ff",
			# activebackground="#5050ff",
		)
		button.pack(side="right")
		###
		frame.pack(fill="x")

	def about_credits_clicked(self):
		about = tix.Toplevel()  # bg="#0f0" does not work
		about.title("Credits")
		about.resizable(False, False)
		set_window_icon(about)
		###
		msg1 = tix.Message(
			about,
			width=500,
			text="\n".join(authors),
			font=("DejaVu Sans", 9, "bold"),
		)
		msg1.pack(fill="x", expand=True)
		###########
		frame = tix.Frame(about)
		closeB = tix.Button(
			frame,
			text="Close",
			command=about.destroy,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		closeB.pack(side="right")
		frame.pack(fill="x")

	def about_license_clicked(self):
		about = tix.Toplevel()  # bg="#0f0" does not work
		about.title("License")
		about.resizable(False, False)
		set_window_icon(about)
		###
		msg1 = tix.Message(
			about,
			width=420,
			text=licenseText,
			font=("DejaVu Sans", 9, "bold"),
		)
		msg1.pack(fill="x", expand=True)
		###########
		frame = tix.Frame(about)
		closeB = tix.Button(
			frame,
			text="Close",
			command=about.destroy,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		closeB.pack(side="right")
		frame.pack(fill="x")

	def quit(self):
		self.rootWin.destroy()

	def resized(self, event):
		dh = self.rootWin.winfo_height() - self.winfo_height()
		# log.debug(dh, self.consoleH)
		# if dh > 20:
		#	self.consoleH += 1
		#	self.console["height"] = self.consoleH
		#	self.console["width"] = int(self.console["width"]) + 1
		#	self.console.grid()
		# for x in dir(self):
		#	if "info" in x:
		#		log.debug(x)

	def combobox_o_changed(self, event):
		# log.debug(self.combobox_o.get())
		formatD = self.combobox_o.get()
		if formatD == noneItem:
			return
		format = Glossary.descFormat[formatD]
		"""
		if format=="Omnidic":
			self.xml.get_widget("label_omnidic_o").show()
			self.xml.get_widget("spinbutton_omnidic_o").show()
		else:
			self.xml.get_widget("label_omnidic_o").hide()
			self.xml.get_widget("spinbutton_omnidic_o").hide()
		if format=="Babylon":
			self.xml.get_widget("label_enc").show()
			self.xml.get_widget("comboentry_enc").show()
		else:
			self.xml.get_widget("label_enc").hide()
			self.xml.get_widget("comboentry_enc").hide()
		"""
		if not self.pref["ui_autoSetOutputFileName"]:  # and format is None:
			return

		pathI = self.entry_i.get()
		pathO = self.entry_o.get()
		formatOD = self.combobox_o.get()

		if formatOD is None:
			return
		if pathO:
			return
		if "." not in pathI:
			return

		extO = Glossary.descExt[formatOD]
		pathO = "".join(os.path.splitext(pathI)[:-1])+extO
		# self.entry_o.delete(0, "end")
		self.entry_o.insert(0, pathO)

	def entry_changed(self, event=None):
		# log.debug("entry_changed")
		# char = event.keysym
		pathI = self.entry_i.get()
		if self.pathI != pathI:
			formatD = self.combobox_i.get()
			if pathI.startswith("file://"):
				pathI = urlToPath(pathI)
				self.entry_i.delete(0, "end")
				self.entry_i.insert(0, pathI)
			if self.pref["ui_autoSetFormat"]:  # format==noneItem:
				ext = os.path.splitext(pathI)[-1].lower()
				if ext in (".gz", ".bz2", ".zip"):
					ext = os.path.splitext(pathI[:-len(ext)])[-1].lower()
				for i in range(len(Glossary.readExt)):
					if ext in Glossary.readExt[i]:
						self.combobox_i.set(Glossary.readDesc[i])
						break
			if self.pref["ui_autoSetOutputFileName"]:  # format==noneItem:
				# pathI = self.entry_i.get()
				formatOD = self.combobox_o.get()
				pathO = self.entry_o.get()
				if formatOD != noneItem and not pathO and "." in pathI:
					extO = Glossary.descExt[formatOD]
					pathO = "".join(os.path.splitext(pathI)[:-1]) + extO
					self.entry_o.delete(0, "end")
					self.entry_o.insert(0, pathO)
			self.pathI = pathI
		##############################################
		pathO = self.entry_o.get()
		if self.pathO != pathO:
			formatD = self.combobox_o.get()
			if pathO.startswith("file://"):
				pathO = urlToPath(pathO)
				self.entry_o.delete(0, "end")
				self.entry_o.insert(0, pathO)
			if self.pref["ui_autoSetFormat"]:  # format==noneItem:
				ext = os.path.splitext(pathO)[-1].lower()
				if ext in (".gz", ".bz2", ".zip"):
					ext = os.path.splitext(pathO[:-len(ext)])[-1].lower()
				for i in range(len(Glossary.writeExt)):
					if ext in Glossary.writeExt[i]:
						self.combobox_o.set(Glossary.writeDesc[i])
						break
			self.pathO = pathO

	def browse_i(self):
		path = filedialog.askopenfilename(initialdir=self.fcd_dir)
		if path:
			self.entry_i.delete(0, "end")
			self.entry_i.insert(0, path)
			self.entry_changed()
			self.fcd_dir = os.path.dirname(path)  # FIXME

	def browse_o(self):
		path = filedialog.asksaveasfilename()
		if path:
			self.entry_o.delete(0, "end")
			self.entry_o.insert(0, path)
			self.entry_changed()
			self.fcd_dir = os.path.dirname(path)  # FIXME

	def convert(self):
		inPath = self.entry_i.get()
		if not inPath:
			log.critical("Input file path is empty!")
			return
		inFormatDesc = self.combobox_i.get()
		if inFormatDesc == noneItem:
			# log.critical("Input format is empty!");return
			inFormat = ""
		else:
			inFormat = Glossary.descFormat[inFormatDesc]

		outPath = self.entry_o.get()
		if not outPath:
			log.critical("Output file path is empty!")
			return
		outFormatDesc = self.combobox_o.get()
		if outFormatDesc in (noneItem, ""):
			log.critical("Output format is empty!")
			return
		outFormat = Glossary.descFormat[outFormatDesc]

		finalOutputFile = self.glos.convert(
			inPath,
			inputFormat=inFormat,
			outputFilename=outPath,
			outputFormat=outFormat,
		)
		# if finalOutputFile:
			# self.status("Convert finished")
		# else:
			# self.status("Convert failed")
		return bool(finalOutputFile)

	def run(self, editPath=None, readOptions=None):
		if readOptions is None:
			readOptions = {}
		# editPath and readOptions are for DB Editor
		# which is not implemented
		self.mainloop()

	def progressInit(self, title):
		self.progressTitle = title

	def progress(self, rat, text=""):
		if not text:
			text = "%%%d" % (rat*100)
		text += " - %s" % self.progressTitle
		self.pbar.updateProgress(rat*100, None, text)
		# self.pbar.value = rat*100
		# self.pbar.update()
		self.rootWin.update()

	def console_clear(self, event=None):
		self.console.delete("1.0", "end")
		self.console.insert("end", "Console:\n")

	def r_browse_i(self):
		pass

	def r_browse_o(self):
		pass

	def r_load(self):
		pass


if __name__ == "__main__":
	import sys
	if len(sys.argv) > 1:
		path = sys.argv[1]
	else:
		path = ""
	ui = UI(path)
	ui.run()



# -*- coding: utf-8 -*-
# Source Glossary for "Sdictionary" (http://sdict.org)
# It has extention '.sdct'

from formats_common import *

enable = True
format = 'SdictSource'
description = 'Sdictionary Source (sdct)'
extentions = ['.sdct']
readOptions = []
writeOptions = [
	'writeInfo',  # bool
	'newline',  # str, or choice ('\r\n', '\n', or '\r')
	'resources',  # bool
]


def write(
	glos,
	filename,
	writeInfo=True,
	newline='\n',
	resources=True,
):
	head = ''
	if writeInfo:
		head += '<header>\n'
		head += 'title = %s\n' % glos.getInfo('name')
		head += 'author = %s\n' % glos.getInfo('author')
		head += 'description = %s\n' % glos.getInfo('description')
		head += 'w_lang = %s\n' % glos.getInfo('inputlang')
		head += 'a_lang = %s\n' % glos.getInfo('outputlang')
		head += '</header>\n#\n#\n#\n'
	glos.writeTxt(
		'___',
		'\n',
		filename,
		writeInfo=False,
		rplList=(
			('\n', '<BR>'),
		),
		ext='.sdct',
		head=head,
		newline=newline,
		resources=resources,
	)



# -*- coding: utf-8 -*-
# ui_qk.py
#
# Copyright © 2010 Saeed Rasooli <saeed.gnu@gmail.com>    (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.


from pyglossary.glossary import *
from .base import *
from os.path import join

from PyQt4 import QtGui as qt
from PyQt4 import QtCore as qc

stderr_saved = sys.stderr
stdout_saved = sys.stdout

# startBold = '\x1b[1m'  # Start Bold # len=4
# startUnderline = '\x1b[4m' # Start Underline # len=4
endFormat = '\x1b[0;0;0m'  # End Format # len=8
# redOnGray = '\x1b[0;1;31;47m'
startRed = '\x1b[31m'

noneItem = 'Not Selected'


class QVirtualFile(object):
	def __init__(self, qtext, mode):
		self.qtext = qtext
		self.mode = mode

	def write(self, text):
		self.qtext.insertPlainText(text)
		if self.mode == 'stdout':
			stdout_saved.write(text)
		elif self.mode == 'stderr':
			stderr_saved.write(startRed+text+endFormat)

	def writelines(self, lines):
		for line in lines:
			self.write(line)

	def flush(self):
		pass

	def isatty(self):
		return 1

	def fileno(self):
		pass


class UI(qt.QWidget, UIBase):
	def __init__(self, ipath, **options):
		qt.QWidget.__init__(self)
		self.setWindowTitle('PyGlossary (Qt)')
		self.setWindowIcon(qt.QIcon(join(uiDir, 'pyglossary.png')))
		######################
		self.running = False
		self.glos = Glossary(ui=self)
		self.pref = {}
		self.pref_load()
		self.pathI = ''
		self.pathO = ''
		self.fcd_dir = join(homeDir, 'Desktop')
		######################
		vbox = qt.QVBoxLayout()
		self.setLayout(vbox)



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2011-2012 kubtek <kubtek@gmail.com>
# This file is part of PyGlossary project, http://github.com/ilius/pyglossary
# Thanks to Raul Fernandes <rgfbr@yahoo.com.br> and Karl Grill
#	   for reverse engineering
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

partOfSpeechByCode = {
	# Use None for codes we have not seen yet
	# Use "" for codes we've seen but part of speech is unknown
	0x30: "noun",
	0x31: "adjective",
	0x32: "verb",
	0x33: "adverb",
	0x34: "interjection",
	0x35: "pronoun",
	0x36: "preposition",
	0x37: "conjunction",
	0x38: "suffix",
	0x39: "prefix",
	0x3A: "article",
	0x3B: "",  # in Babylon Italian-English.BGL,
	# Babylon Spanish-English.BGL,
	# Babylon_Chinese_S_English.BGL
	# no indication of the part of speech
	0x3C: "abbreviation",
	# (short form: 'ר"ת')
	# (full form: "ר"ת: ראשי תיבות")
	# "ת'"
	# adjective
	# (full form: "ת': תואר")

	# "ש"ע"
	# noun
	# (full form: "ש"ע: שם עצם")

	0x3D: "masculine noun and adjective",
	0x3E: "feminine noun and adjective",
	0x3F: "masculine and feminine noun and adjective",
	0x40: "feminine noun",
	# (short form: "נ\'")
	# (full form: "נ': נקבה")
	0x41: "masculine and feminine noun",
	# 0x41: noun that may be used as masculine and feminine
	# (short form: "זו"נ")
	# (full form: "זו"נ: זכר ונקבה")
	0x42: "masculine noun",
	# (short form: 'ז\'')
	# (full form: "ז': זכר")
	0x43: "numeral",
	0x44: "participle",
	0x45: None,
	0x46: None,
	0x47: None,
}



from pyglossary.file_utils import fileCountLines
from pyglossary.entry import Entry

import logging
log = logging.getLogger('root')


class TextGlossaryReader(object):
	def __init__(self, glos, hasInfo=True):
		self._glos = glos
		self._filename = ''
		self._file = None
		self._hasInfo = True
		self._leadingLinesCount = 0
		self._pendingEntries = []
		self._wordCount = None
		self._pos = -1

	def open(self, filename, encoding='utf-8'):
		self._filename = filename
		self._file = open(filename, 'r', encoding=encoding)
		if self._hasInfo:
			self.loadInfo()

	def close(self):
		if not self._file:
			return
		try:
			self._file.close()
		except:
			log.exception('error while closing file "%s"' % self._filename)
		self._file = None

	def loadInfo(self):
		self._pendingEntries = []
		self._leadingLinesCount = 0
		try:
			while True:
				wordDefi = self.nextPair()
				if not wordDefi:
					continue
				word, defi = wordDefi
				if not self.isInfoWord(word):
					self._pendingEntries.append(Entry(word, defi))
					break
				self._leadingLinesCount += 1
				word = self.fixInfoWord(word)
				if not word:
					continue
				self._glos.setInfo(word, defi)
		except StopIteration:
			pass

	def __next__(self):
		self._pos += 1
		try:
			return self._pendingEntries.pop(0)
		except IndexError:
			pass
		###
		try:
			wordDefi = self.nextPair()
		except StopIteration as e:
			self._wordCount = self._pos
			raise e
		if not wordDefi:
			return
		word, defi = wordDefi
		###
		return Entry(word, defi)

	def __len__(self):
		if self._wordCount is None:
			log.debug('Try not to use len(reader) as it takes extra time')
			self._wordCount = fileCountLines(self._filename) - \
				self._leadingLinesCount
		return self._wordCount

	def __iter__(self):
		return self

	def isInfoWord(self, word):
		raise NotImplementedError

	def fixInfoWord(self, word):
		raise NotImplementedError

	def nextPair(self):
		raise NotImplementedError



# -*- coding: utf-8 -*-

from formats_common import *
from pyglossary.text_reader import TextGlossaryReader
from pyglossary.text_utils import escapeNTB, unescapeNTB, splitByBarUnescapeNTB

enable = True
format = "Tabfile"
description = "Tabfile (txt, dic)"
extentions = [".txt", ".tab", ".dic"]
readOptions = [
	"encoding",
]
writeOptions = [
	"encoding",  # str
	"writeInfo",  # bool
	"resources",  # bool
]


class Reader(TextGlossaryReader):
	def isInfoWord(self, word):
		if isinstance(word, str):
			return word.startswith("#")
		else:
			return False

	def fixInfoWord(self, word):
		if isinstance(word, str):
			return word.lstrip("#")
		else:
			return word

	def nextPair(self):
		if not self._file:
			raise StopIteration
		line = self._file.readline()
		if not line:
			raise StopIteration
		line = line.strip()  # This also removes tailing newline
		if not line:
			return
		###
		word, tab, defi = line.partition("\t")
		if not tab:
			log.error(
				"Warning: line starting with \"%s\" has no tab!" % line[:10]
			)
			return
		###
		if self._glos.getPref("enable_alts", True):
			word = splitByBarUnescapeNTB(word)
			if len(word) == 1:
				word = word[0]
		else:
			word = unescapeNTB(word, bar=True)
		###
		defi = unescapeNTB(defi)
		###
		return word, defi


def write(
	glos,
	filename,
	encoding="utf-8",
	writeInfo=True,
	resources=True,
):
	return glos.writeTabfile(
		filename,
		encoding=encoding,
		writeInfo=writeInfo,
		resources=resources,
	)



# -*- coding: utf-8 -*-
import re
from tempfile import mktemp
import os
from os.path import (
	join,
	exists,
	dirname,
)


class DataEntry(object): # or Resource? FIXME
	def isData(self):
		return True

	def __init__(self, fname, data, inTmp=False):
		assert isinstance(fname, str)
		assert isinstance(data, bytes)
		assert isinstance(inTmp, bool)

		if inTmp:
			tmpPath = mktemp(prefix=fname + "_")
			with open(tmpPath, "wb") as toFile:
				toFile.write(data)
			data = ""
		else:
			tmpPath = None

		self._fname = fname
		self._data = data  # bytes instance
		self._tmpPath = tmpPath

	def getFileName(self):
		return self._fname

	def getData(self):
		if self._tmpPath:
			with open(self._tmpPath, "rb") as fromFile:
				return fromFile.read()
		else:
			return self._data

	def save(self, directory):
		fname = self._fname
		# fix filename depending on operating system? FIXME
		fpath = join(directory, fname)
		fdir = dirname(fpath)
		if not exists(fdir):
			os.makedirs(fdir)
		with open(fpath, "wb") as toFile:
			toFile.write(self.getData())
		return fpath

	def getWord(self):
		return self._fname

	def getWords(self):
		return [self._fname]

	def getDefi(self):
		return "File: %s" % self._fname  

	def getDefis(self):
		return [self.getDefi()]

	def getDefiFormat(self):
		return "b" # "m" or "b" (binary) FIXME

	def setDefiFormat(self, defiFormat):
		pass

	def detectDefiFormat(self):
		pass

	def addAlt(self, alt):
		pass

	def editFuncWord(self, func):
		pass
		# modify fname?
		# FIXME

	def editFuncDefi(self, func):
		pass

	def strip(self):
		pass

	def replaceInWord(self, source, target):
		pass

	def replaceInDefi(self, source, target):
		pass

	def replace(self, source, target):
		pass

	def getRaw(self):
		return (
			self._fname,
			"DATA",
			self,
		)


class Entry(object):
	sep = "|"
	htmlPattern = re.compile(
		".*(" + "|".join([
			r"<br\s*/?\s*>",
			r"<p[ >]",
			r"<div[ >]",
			r"<a href=",
			r"<sup[ >]",
		]) + ")",
		re.S,
	)

	def isData(self):
		return False

	def _join(self, parts):
		return self.sep.join([
			part.replace(self.sep, "\\"+self.sep)
			for part in parts
		])

	@staticmethod
	def getEntrySortKey(key=None):
		if key:
			return lambda entry: key(entry.getWords()[0])
		else:
			return lambda entry: entry.getWords()[0]

	@staticmethod
	def getRawEntrySortKey(key=None):
		if key:
			return lambda x: key(
				x[0][0] if isinstance(x[0], (list, tuple)) else x[0]
			)
		else:
			return lambda x: \
				x[0][0] if isinstance(x[0], (list, tuple)) else x[0]

	def __init__(self, word, defi, defiFormat="m"):
		"""
			word: string or a list of strings (including alternate words)
			defi: string or a list of strings (including alternate definitions)
			defiFormat (optional): definition format:
				"m": plain text
				"h": html
				"x": xdxf
		"""

		# memory optimization:
		if isinstance(word, list):
			if len(word) == 1:
				word = word[0]
		elif not isinstance(word, str):
			raise TypeError("invalid word type %s" % type(word))

		if isinstance(defi, list):
			if len(defi) == 1:
				defi = defi[0]
		elif not isinstance(defi, str):
			raise TypeError("invalid defi type %s" % type(defi))

		if not defiFormat in ("m", "h", "x"):
			raise ValueError("invalid defiFormat %r" % defiFormat)

		self._word = word
		self._defi = defi
		self._defiFormat = defiFormat

	def getWord(self):
		"""
			returns string of word,
				and all the alternate words
				seperated by "|"
		"""
		if isinstance(self._word, str):
			return self._word
		else:
			return self._join(self._word)

	def getWords(self):
		"""
			returns list of the word and all the alternate words
		"""
		if isinstance(self._word, str):
			return [self._word]
		else:
			return self._word

	def getDefi(self):
		"""
			returns string of definition,
				and all the alternate definitions
				seperated by "|"
		"""
		if isinstance(self._defi, str):
			return self._defi
		else:
			return self._join(self._defi)

	def getDefis(self):
		"""
			returns list of the definition and all the alternate definitions
		"""
		if isinstance(self._defi, str):
			return [self._defi]
		else:
			return self._defi

	def getDefiFormat(self):
		"""
			returns definition format:
				"m": plain text
				"h": html
				"x": xdxf
		"""
		return self._defiFormat

	def setDefiFormat(self, defiFormat):
		"""
			defiFormat:
				"m": plain text
				"h": html
				"x": xdxf
		"""
		self._defiFormat = defiFormat

	def detectDefiFormat(self):
		if self._defiFormat != "m":
			return
		defi = self.getDefi().lower()
		if re.match(self.htmlPattern, defi):
			self._defiFormat = "h"

	def addAlt(self, alt):
		words = self.getWords()
		words.append(alt)
		self._word = words

	def editFuncWord(self, func):
		"""
			run function `func` on all the words
			`func` must accept only one string as argument
			and return the modified string
		"""
		if isinstance(self._word, str):
			self._word = func(self._word)
		else:
			self._word = tuple(
				func(st) for st in self._word
			)

	def editFuncDefi(self, func):
		"""
			run function `func` on all the definitions
			`func` must accept only one string as argument
			and return the modified string
		"""
		if isinstance(self._defi, str):
			self._defi = func(self._defi)
		else:
			self._defi = tuple(
				func(st) for st in self._defi
			)

	def strip(self):
		"""
			strip whitespaces from all words and definitions
		"""
		self.editFuncWord(str.strip)
		self.editFuncDefi(str.strip)

	def replaceInWord(self, source, target):
		"""
			replace string `source` with `target` in all words
		"""
		if isinstance(self._word, str):
			self._word = self._word.replace(source, target)
		else:
			self._word = tuple(
				st.replace(source, target) for st in self._word
			)

	def replaceInDefi(self, source, target):
		"""
			replace string `source` with `target` in all definitions
		"""
		if isinstance(self._defi, str):
			self._defi = self._defi.replace(source, target)
		else:
			self._defi = tuple(
				st.replace(source, target) for st in self._defi
			)

	def replace(self, source, target):
		"""
			replace string `source` with `target` in all words and definitions
		"""
		self.replaceInWord(source, target)
		self.replaceInDefi(source, target)

	def getRaw(self):
		"""
			returns a tuple (word, defi) or (word, defi, defiFormat)
			where both word and defi might be string or list of strings
		"""
		if self._defiFormat:
			return (
				self._word,
				self._defi,
				self._defiFormat,
			)
		else:
			return (
				self._word,
				self._defi,
			)

	@classmethod
	def fromRaw(cls, rawEntry, defaultDefiFormat="m"):
		"""
			rawEntry can be (word, defi) or (word, defi, defiFormat)
			where both word and defi can be string or list of strings
			if defiFormat is missing, defaultDefiFormat will be used

			creates and return an Entry object from `rawEntry` tuple
		"""
		word = rawEntry[0]
		defi = rawEntry[1]
		if defi == "DATA":
			try:
				dataEntry = rawEntry[2] # DataEntry instance
			except IndexError:
				pass
			else:
				# if isinstance(dataEntry, DataEntry)  # FIXME
				return dataEntry
		try:
			defiFormat = rawEntry[2]
		except IndexError:
			defiFormat = defaultDefiFormat

		if isinstance(word, tuple):
			word = list(word)
		if isinstance(defi, tuple):
			defi = list(defi)

		return cls(
			word,
			defi,
			defiFormat=defiFormat,
		)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ui_main.py
#
# Copyright © 2008-2010 Saeed Rasooli <saeed.gnu@gmail.com>    (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import os
import sys
import argparse
import builtins
from os.path import dirname, join, realpath
from pprint import pformat
import logging

from pyglossary import core  # essential
from pyglossary import VERSION
from pyglossary.text_utils import startRed, endFormat

# the first thing to do is to set up logger.
# other modules also using logger 'root', so it's essential to set it up prior
# to importing anything else; with exception to pyglossary.core which sets up
# logger class, and so should be done before actually initializing logger.
# verbosity level may be given on command line, so we have to parse arguments
# before setting up logger.
# once more:
# - import system modules like os, sys, argparse etc and pyglossary.core
# - parse args
# - set up logger
# - import submodules
# - other code

# no-progress-bar only for command line UI
# FIXME: load ui-dependent available options from ui modules
# (for example ui_cmd.available_options)
# the only problem is that it has to "import gtk" before it get the
# "ui_gtk.available_options"

# FIXME
# -v (verbose or version?)
# -r (reverse or read-options)

parser = argparse.ArgumentParser(add_help=False)

parser.add_argument(
	'-v',
	'--verbosity',
	action='store',
	dest='verbosity',
	type=int,
	choices=(0, 1, 2, 3, 4),
	required=False,
	default=3,  # FIXME
)
parser.add_argument(
	'--version',
	action='version',
	version='PyGlossary %s' % VERSION,
)
parser.add_argument(
	'-h',
	'--help',
	dest='help',
	action='store_true',
)
parser.add_argument(
	'-u',
	'--ui',
	dest='ui_type',
	default='auto',
	choices=(
		'cmd',
		'gtk',
		'tk',
		# 'qt',
		'auto',
		'none',
	),
)
parser.add_argument(
	'-r',
	'--read-options',
	dest='readOptions',
	default='',
)
parser.add_argument(
	'-w',
	'--write-options',
	dest='writeOptions',
	default='',
)
parser.add_argument(
	'--read-format',
	dest='inputFormat',
)
parser.add_argument(
	'--write-format',
	dest='outputFormat',
	action='store',
)
parser.add_argument(
	'--direct',
	dest='direct',
	action='store_true',
	default=None,
	help='if possible, convert directly without loading into memory',
)
parser.add_argument(
	'--indirect',
	dest='direct',
	action='store_false',
	default=None,
	help='disable `direct` mode, load full data into memory before writing'
		 ', this is default',
)
parser.add_argument(
	'--reverse',
	dest='reverse',
	action='store_true',
)
parser.add_argument(
	'--no-progress-bar',
	dest='progressbar',
	action='store_false',
	default=None,
)
parser.add_argument(
	'--sort',
	dest='sort',
	action='store_true',
	default=None,
)
parser.add_argument(
	'--no-sort',
	dest='sort',
	action='store_false',
	default=None,
)
parser.add_argument(
	'--sort-cache-size',
	dest='sortCacheSize',
	type=int,
	default=None,
)

parser.add_argument(
	'--utf8-check',
	dest='utf8Check',
	action='store_true',
	default=None,
)
parser.add_argument(
	'--no-utf8-check',
	dest='utf8Check',
	action='store_false',
	default=None,
)
parser.add_argument(
	'--lower',
	dest='lower',
	action='store_true',
	default=None,
	help='lowercase words before writing',
)
parser.add_argument(
	'--no-lower',
	dest='lower',
	action='store_false',
	default=None,
	help='don\'t lowercase words before writing',
)
parser.add_argument(
	'--skip-resources',
	dest='skipResources',
	action='store_true',
	default=None,
	help='skip resources (images, audio, etc)',
)

parser.add_argument(
	'--no-color',
	dest='noColor',
	action='store_true',
)
parser.add_argument(
	'inputFilename',
	action='store',
	default='',
	nargs='?',
)
parser.add_argument(
	'outputFilename',
	action='store',
	default='',
	nargs='?',
)

args = parser.parse_args()

log = logging.getLogger('root')
log.setVerbosity(args.verbosity)
log.addHandler(
	core.StdLogHandler(noColor=args.noColor),
)
# with the logger setted up, we can import other pyglossary modules, so they
# can do some loggging in right way.


core.checkCreateConfDir()


##############################

from pyglossary.glossary import Glossary
from ui.ui_cmd import COMMAND, help, parseFormatOptionsStr

##############################


def dashToCamelCase(text):  # converts "hello-PYTHON-user" to "helloPythonUser"
	parts = text.split('-')
	parts[0] = parts[0].lower()
	for i in range(1, len(parts)):
		parts[i] = parts[i].capitalize()
	return ''.join(parts)

ui_list = (
	'gtk',
	'tk',
	'qt',
)

# log.info('PyGlossary %s'%VERSION)


if args.help:
	help()
	sys.exit(0)


if os.sep != '/':
	args.noColor = True


# only used in ui_cmd for now
readOptions = parseFormatOptionsStr(args.readOptions)
writeOptions = parseFormatOptionsStr(args.writeOptions)


"""
	examples for read and write options:
	--read-options testOption=stringValue
	--read-options enableFoo=True
	--read-options fooList=[1,2,3]
	--read-options 'fooList=[1, 2, 3]'
	--read-options 'testOption=stringValue; enableFoo=True; fooList=[1, 2, 3]'
	--read-options 'testOption=stringValue;enableFoo=True;fooList=[1,2,3]'
"""

# FIXME
prefOptionsKeys = (
	# 'verbosity',
	'utf8Check',
	'lower',
	'skipResources',
)

convertOptionsKeys = (
	'direct',
	'progressbar',
	'sort',
	'sortCacheSize',
	# 'sortKey',# or sortAlg FIXME
)

prefOptions = {}
for param in prefOptionsKeys:
	value = getattr(args, param, None)
	if value is not None:
		prefOptions[param] = value

convertOptions = {}
for param in convertOptionsKeys:
	value = getattr(args, param, None)
	if value is not None:
		convertOptions[param] = value

log.pretty(prefOptions, 'prefOptions = ')
log.pretty(readOptions, 'readOptions = ')
log.pretty(writeOptions, 'writeOptions = ')
log.pretty(convertOptions, 'convertOptions = ')

"""
ui_type: User interface type
Possible values:
	cmd - Command line interface, this ui will automatically selected
		  if you give both input and output file
	gtk - GTK interface
	tk - Tkinter interface
	qt - Qt interface
	auto - Use the first available UI
"""
ui_type = args.ui_type


if args.inputFilename:
	if args.outputFilename and ui_type != 'none':
		ui_type = 'cmd'  # silently? FIXME
else:
	if ui_type == 'cmd':
		log.error('no input file given, try --help')
		exit(1)

if ui_type == 'none':
	if args.reverse:
		log.error('--reverse does not work with --ui=none')
		sys.exit(1)
	glos = Glossary()
	glos.convert(
		args.inputFilename,
		inputFormat=args.inputFormat,
		outputFilename=args.outputFilename,
		outputFormat=args.outputFormat,
		readOptions=readOptions,
		writeOptions=writeOptions,
		**convertOptions
	)
	sys.exit(0)
elif ui_type == 'cmd':
	from ui import ui_cmd
	sys.exit(0 if ui_cmd.UI().run(
		args.inputFilename,
		outputFilename=args.outputFilename,
		inputFormat=args.inputFormat,
		outputFormat=args.outputFormat,
		reverse=args.reverse,
		prefOptions=prefOptions,
		readOptions=readOptions,
		writeOptions=writeOptions,
		convertOptions=convertOptions,
	) else 1)
if ui_type == 'auto':
	ui_module = None
	for ui_type2 in ui_list:
		try:
			ui_module = getattr(
				__import__('ui.ui_%s' % ui_type2),
				'ui_%s' % ui_type2,
			)
		except ImportError:
			log.exception('error while importing UI module:')  # FIXME
		else:
			break
	if ui_module is None:
		log.error('no user interface module found! try "%s -h" to see command line usage' % sys.argv[0])
		sys.exit(1)
else:
	ui_module = getattr(
		__import__('ui.ui_%s' % ui_type),
		'ui_%s' % ui_type,
	)

sys.exit(0 if ui_module.UI(**prefOptions).run(
	editPath=args.inputFilename,
	readOptions=readOptions,
) else 1)



# -*- coding: utf-8 -*-
# ui_cmd.py
#
# Copyright © 2008-2010 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from os.path import join
import time
import signal

from pyglossary.glossary import *
from .base import *
from . import progressbar as pb


if os.sep == "\\":  # Operating system is Windows
	startRed = ""
	endFormat = ""
	startBold = ""
	startUnderline = ""
	endFormat = ""
else:
	startRed = "\x1b[31m"
	endFormat = "\x1b[0;0;0m"

	startBold = "\x1b[1m"  # Start Bold # len=4
	startUnderline = "\x1b[4m"  # Start Underline # len=4
	endFormat = "\x1b[0;0;0m"  # End Format # len=8
	# redOnGray = "\x1b[0;1;31;47m"


COMMAND = "pyglossary"
# COMMAND = sys.argv[0]


def getColWidth(subject, strings):
	return max(
		len(x) for x in [subject] + strings
	)


def getFormatsTable(names, header):
	descriptions = [
		Glossary.formatsDesc[name]
		for name in names
	]
	extentions = [
		" ".join(Glossary.formatsExt[name])
		for name in names
	]

	nameWidth = getColWidth("Name", names)
	descriptionWidth = getColWidth("Description", descriptions)
	extentionsWidth = getColWidth("Extentions", extentions)

	lines = ["\n"]
	lines.append("%s%s%s" % (startBold, header, endFormat))

	lines.append(
		" | ".join([
			"Name".center(nameWidth),
			"Description".center(descriptionWidth),
			"Extentions".center(extentionsWidth)
		])
	)
	lines.append(
		"-+-".join([
			"-" * nameWidth,
			"-" * descriptionWidth,
			"-" * extentionsWidth,
		])
	)
	for index, name in enumerate(names):
		lines.append(
			" | ".join([
				name.ljust(nameWidth),
				descriptions[index].ljust(descriptionWidth),
				extentions[index].ljust(extentionsWidth)
			])
		)

	return "\n".join(lines)


def help():
	import string
	with open(join(dataDir, "help")) as fp:
		text = fp.read()
	text = text.replace("<b>", startBold)\
		.replace("<u>", startUnderline)\
		.replace("</b>", endFormat)\
		.replace("</u>", endFormat)
	text = string.Template(text).substitute(
		CMD=COMMAND,
	)
	text += getFormatsTable(Glossary.readFormats, "Supported input formats:")
	text += getFormatsTable(Glossary.writeFormats, "Supported output formats:")
	print(text)


def parseFormatOptionsStr(st):
	st = st.strip()
	if not st:
		return {}

	opt = {}
	parts = st.split(";")
	for part in parts:
		try:
			(key, value) = part.split("=")
		except ValueError:
			log.error("bad option syntax: %s" % part)
			continue
		key = key.strip()
		value = value.strip()
		# if it is string form of a number or boolean or tuple ...
		try:
			newValue = eval(value)
		except:
			pass
		else:
			if isinstance(newValue, (
				bool,
				int,
				float,
				tuple,
				list,
				dict,
			)):
				value = newValue
		opt[key] = value
	return opt


class NullObj(object):
	def __getattr__(self, attr):
		return self

	def __setattr__(self, attr, value):
		pass

	def __call__(self, *args, **kwargs):
		pass


class UI(UIBase):
	def __init__(self, **options):
		self.pref = {}
		# log.debug(self.pref)
		self.pbar = NullObj()
		self._toPause = False

	def onSigInt(self, *args):
		if self._toPause:
			log.info("\nOperation Canceled")
			sys.exit(0)
		else:
			self._toPause = True
			log.info("\nPlease wait...")

	def setText(self, text):
		self.pbar.widgets[0] = text

	def progressInit(self, title):
		rot = pb.RotatingMarker()
		self.pbar = pb.ProgressBar(
			widgets=[
				title,
				pb.Bar(marker="█", right=rot),
				pb.Percentage(),
				"% ",
				pb.ETA(),
			],
			maxval=1.0,
			update_step=0.5,
		)
		rot.pbar = self.pbar

	def progress(self, rat, text=""):
		self.pbar.update(rat)

	def reverseLoop(self, *args, **kwargs):
		reverseKwArgs = {}
		for key in (
			"words",
			"matchWord",
			"showRel",
			"includeDefs",
			"reportStep",
			"saveStep",
			"maxNum",
			"minRel",
			"minWordLen"
		):
			try:
				reverseKwArgs[key] = self.pref["reverse_" + key]
			except KeyError:
				pass
		reverseKwArgs.update(kwargs)

		if not self._toPause:
			log.info("Reversing glossary... (Press Ctrl+C to pause/stop)")
		for wordI in self.glos.reverse(**reverseKwArgs):
			if self._toPause:
				log.info(
					"Reverse is paused."
					" Press Enter to continue, and Ctrl+C to exit"
				)
				input()
				self._toPause = False

	def run(
		self,
		inputFilename,
		outputFilename="",
		inputFormat="",
		outputFormat="",
		reverse=False,
		prefOptions=None,
		readOptions=None,
		writeOptions=None,
		convertOptions=None,
	):
		if not prefOptions:
			prefOptions = {}
		if not readOptions:
			readOptions = {}
		if not writeOptions:
			writeOptions = {}
		if not convertOptions:
			convertOptions = {}

		self.pref_load(**prefOptions)

		if inputFormat:
			# inputFormat = inputFormat.capitalize()
			if inputFormat not in Glossary.readFormats:
				log.error("invalid read format %s" % inputFormat)
		if outputFormat:
			# outputFormat = outputFormat.capitalize()
			if outputFormat not in Glossary.writeFormats:
				log.error("invalid write format %s" % outputFormat)
				log.error("try: %s --help" % COMMAND)
				return 1
		if not outputFilename:
			if reverse:
				pass
			elif outputFormat:
				try:
					ext = Glossary.formatsExt[outputFormat][0]
				except (KeyError, IndexError):
					log.error("invalid write format %s" % outputFormat)
					log.error("try: %s --help" % COMMAND)
					return 1
				else:
					outputFilename = os.path.splitext(inputFilename)[0] + ext
			else:
				log.error("neither output file nor output format is given")
				log.error("try: %s --help" % COMMAND)
				return 1

		glos = self.glos = Glossary(ui=self)
		if reverse:
			signal.signal(signal.SIGINT, self.onSigInt)  # good place? FIXME
			readOptions["direct"] = False
			if not glos.read(
				inputFilename,
				format=inputFormat,
				**readOptions
			):
				log.error("reading input file was failed!")
				return False
			self.setText("Reversing: ")
			self.pbar.update_step = 0.1
			self.reverseLoop(savePath=outputFilename)
		else:
			finalOutputFile = self.glos.convert(
				inputFilename,
				inputFormat=inputFormat,
				outputFilename=outputFilename,
				outputFormat=outputFormat,
				readOptions=readOptions,
				writeOptions=writeOptions,
				**convertOptions
			)
			return bool(finalOutputFile)

		return True



# -*- coding: utf-8 -*-
# flawless_dsl/main.py
#
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
"""
exposed API lives here.
"""


import copy
import re

from . import tag as _tag
from . import layer as _layer


def process_closing_tags(stack, tags):
	"""
	close `tags`, closing some inner layers if necessary.

	:param stack: Iterable[layer.Layer]
	:param tags: Iterable[str]
	"""
	index = len(stack) - 1
	for tag in copy.copy(tags):
		index_for_tag = _tag.index_of_layer_containing_tag(stack, tag)
		if index_for_tag is not None:
			index = min(index, index_for_tag)
		else:
			tags.remove(tag)

	if not tags:
		return

	to_open = set()
	for layer in stack[:index:-1]:
		for lt in layer.tags:
			if lt.closing not in tags:
				to_open.add(lt)
		_layer.close_layer(stack)

	to_close = set()
	layer = stack[index]
	for lt in layer.tags:
		if lt.closing in tags:
			to_close.add(lt)
	_layer.close_tags(stack, to_close, index)

	if to_open:
		_layer.Layer(stack)
		stack[-1].tags = to_open


OPEN = 1
CLOSE = 2
TEXT = 3

BRACKET_L = "\0\1"
BRACKET_R = "\0\2"

# precompiled regexs
re_m_tag_with_content = re.compile(r"(\[m\d\])(.*?)(\[/m\])")
re_non_escaped_bracket = re.compile(r"(?<!\\)\[")
_startswith_tag_cache = {}


class FlawlessDSLParser(object):
	"""
	only clean flawless dsl on output!
	"""

	def __init__(self, tags=frozenset({
		("m", "\d"),
		"*",
		"ex",
		"i",
		("c", "(?: \w+)?"),
		"p",
		"\"",
		"b",
		"s",
		"sup",
		"sub",
		"ref",
		"url",
	})):
		"""
		:type tags: set[str | tuple[str]] | frozenset[str | tuple[str]]
		:param tags: set (or any other iterable) of tags where each tag is a
					 string or two-tuple.  if string, it is tag name without
					 brackets, must be constant, i.e. non-save regex characters
					 will be escaped, e.g.: "i", "sub", "*".
					 if 2-tuple, then first item is tag"s base name, and
					 second is its extension for opening tag,
					 e.g.: ("c", r" (\w+)"), ("m", r"\d")
		"""
		tags_ = set()
		for tag, ext_re in (
			t if isinstance(t, tuple) else (t, "")
			for t in tags
		):
			tag_re = re.escape(tag)
			tag_open_re = r"\[%s%s\]" % (tag_re, ext_re)
			tags_.add((tag, tag_re, ext_re, tag_open_re))
		self.tags = frozenset(tags_)

	def parse(self, line):
		r"""
		parse dsl markup in `line` and return clean valid dsl markup.

		:type line: str
		:param line: line with dsl formatting.

		:rtype: str
		"""
		line = self.put_brackets_away(line)
		line = self._parse(line)
		return self.bring_brackets_back(line)

	def _parse(self, line):
		items = self._split_line_by_tags(line)
		line = self._tags_and_text_loop(items)
		return line

	def _split_line_by_tags(self, line):
		"""
		split line into chunks, each chunk is whether opening / closing
		tag or text.

		return iterable of two-tuples. first element is item's type, one of:
		- OPEN, second element is Tag object
		- CLOSE, second element is str with closed tag's name
		- TEXT, second element is str

		:param line: str
		:return: Iterable
		"""
		ptr = 0
		while ptr < len(line):
			bracket = line.find("[", ptr)
			if bracket != -1:
				chunk = line[ptr:bracket]
			else:
				chunk = line[ptr:]

			if chunk:
				yield TEXT, chunk

			if bracket == -1:
				break

			ptr = bracket
			# at least two chars after opening bracket:
			bracket = line.find("]", ptr + 2)
			if line[ptr + 1] == "/":
				yield CLOSE, line[ptr + 2:bracket]
			else:
				for tag, _, _, tag_open_re in self.tags:
					if re.match(tag_open_re, line[ptr:bracket + 1]):
						yield OPEN, _tag.Tag(line[ptr + 1:bracket], tag)
						break
				else:
					tag = line[ptr + 1:bracket]
					yield OPEN, _tag.Tag(tag, tag)
			ptr = bracket + 1

	@staticmethod
	def _tags_and_text_loop(tags_and_text):
		"""
		parse chunks one by one.

		:param tags_and_text:
			Iterable[["OPEN", Tag] | ["CLOSE", str] | ["TEXT", str]]
		:return: str
		"""
		state = TEXT
		stack = []
		closings = set()

		for item_t, item in tags_and_text:

			if item_t is OPEN:
				if _tag.was_opened(stack, item) and \
				  item.closing not in closings:
					continue

				if item.closing == "m" and len(stack) >= 1:
					# close all layers.  [m*] tags can only appear
					# at top layer.
					# note: do not reopen tags that were marked as
					# closed already.
					to_open = set.union(*(
						{t for t in l.tags if t.closing not in closings}
						for l in stack
					))
					for i in range(len(stack)):
						_layer.close_layer(stack)
					# assert len(stack) == 1
					# assert not stack[0].tags
					_layer.Layer(stack)
					stack[-1].tags = to_open

				elif state is CLOSE:
					process_closing_tags(stack, closings)

				if not stack or stack[-1].text:
					_layer.Layer(stack)

				stack[-1].tags.add(item)
				state = OPEN
				continue

			elif item_t is CLOSE:
				if state in (OPEN, TEXT):
					closings.clear()
				closings.add(item)
				state = CLOSE
				continue

			elif item_t is TEXT:
				if state is CLOSE:
					process_closing_tags(stack, closings)

				if not stack:
					_layer.Layer(stack)
				stack[-1].text += item
				state = TEXT
				continue

		if state is CLOSE and closings:
			process_closing_tags(stack, closings)
		# shutdown unclosed tags
		return "".join([l.text for l in stack])

	def put_brackets_away(self, line):
		"""put away \[, \] and brackets that does not belong to any of given tags.

		:rtype: str
		"""
		clean_line = ""
		startswith_tag = _startswith_tag_cache.get(self.tags, None)
		if startswith_tag is None:
			openings = "|".join("%s%s" % (_[1], _[2]) for _ in self.tags)
			closings = "|".join(_[1] for _ in self.tags)
			startswith_tag = re.compile(
				r"(?:(?:%s)|/(?:%s))\]" % (openings, closings)
			)
			_startswith_tag_cache[self.tags] = startswith_tag
		for i, chunk in enumerate(re_non_escaped_bracket.split(line)):
			if i != 0:
				m = startswith_tag.match(chunk)
				if m:
					clean_line += "[%s%s" % (
						m.group(),
						chunk[m.end():].replace("[", BRACKET_L)
						.replace("]", BRACKET_R)
					)
				else:
					clean_line += BRACKET_L + chunk.replace("[", BRACKET_L)\
						.replace("]", BRACKET_R)
			else:  # firsr chunk
				clean_line += chunk.replace("[", BRACKET_L)\
					.replace("]", BRACKET_R)
		return clean_line

	@staticmethod
	def bring_brackets_back(line):
		return line.replace(BRACKET_L, "[").replace(BRACKET_R, "]")


def parse(line, tags=None):
	"""parse DSL markup.

	WARNING!
	`parse` function is not optimal because it creates new parser instance
	on each call.
	consider cache one [per thread] instance of FlawlessDSLParser in your code.
	"""
	import warnings
	warnings.warn(
		"`parse` function is not optimal because it creates new parser "
		"instance on each call.\n"
		"consider cache one [per thread] instance of FlawlessDSLParser "
		"in your code."
	)
	if tags:
		parser = FlawlessDSLParser(tags)
	else:
		parser = FlawlessDSLParser()
	return parser.parse(line)



# -*- coding: utf-8 -*-

from formats_common import *

enable = False
format = 'Test'
description = 'Test Format File(.test)'
extentions = ['.test', '.tst']
readOptions = []
writeOptions = []


def read(glos, filename):  # glos is a Glossary object, filename is a string
	log.info('reading from format %s using plugin' % format)
	count = 100
	# get number of entries from input file(depending on your format)
	for i in range(count):
		# here get word and definition from file(depending on your format)
		word = 'word_%s' % i
		defi = 'definition %s' % i
		glos.addEntry(word, defi)
	# here read info from file and set to Glossary object
	glos.setInfo('name', 'Test')
	glos.setInfo('descriptin', 'Test glossary craeted by a PyGlossary plugin')
	glos.setInfo('author', 'Me')
	glos.setInfo('copyright', 'GPL')
	return True  # reading input file was succesfull


def write(glos, filename):  # glos is a Glossary object, filename is a string
	log.info('writing to format %s using plugin' % format)
	for entry in glos:
		word = entry.getWord()
		defi = entry.getDefi()
		# here write word and defi to the output file (depending on
		# your format)
	# here read info from Glossaey object
	name = glos.getInfo('name')
	descriptin = glos.getInfo('descriptin')
	author = glos.getInfo('author')
	copyright = glos.getInfo('copyright')
	# if an info key doesn't exist, getInfo returns empty string
	# now write info to the output file (depending on your output format)
	return True  # writing output file was succesfull



# -*- coding: utf-8 -*-
#
# Copyright (C) Saeed Rasooli <saeed.gnu@gmail.com>
# Copyright (C) 2007 Mehdi Bayazee <Bayazee@Gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/gpl.txt>.
# Also avalable in /usr/share/common-licenses/GPL on Debian systems
# or /usr/share/licenses/common/GPL3/license.txt on ArchLinux


# Gregorian calendar:
#   http://en.wikipedia.org/wiki/Gregorian_calendar

from datetime import datetime

name = 'gregorian'
desc = 'Gregorian'
origLang = 'en'

monthName = ('January', 'February', 'March', 'April', 'May', 'June',
			 'July', 'August', 'September', 'October', 'November', 'December')

monthNameAb = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
			   'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')


def getMonthName(m, y=None):
	return monthName.__getitem__(m - 1)


def getMonthNameAb(m, y=None):
	return monthNameAb.__getitem__(m - 1)


def getMonthsInYear(y):
	return 12


epoch = 1721426
minMonthLen = 29
maxMonthLen = 31
avgYearLen = 365.2425  # FIXME

options = ()


def save():
	pass


def isLeap(y):
	if y < 1:
		y += 1
	return y % 4 == 0 and not (y % 100 == 0 and y % 400 != 0)


def to_jd(year, month, day):
	if year > 0:  # > 1.5x faster
		return datetime(year, month, day).toordinal() + 1721425

	# Python 2.x and 3.x:
	if month <= 2:
		tm = 0
	elif isLeap(year):
		tm = -1
	else:
		tm = -2
	# Python >= 2.5:
	# tm = 0 if month <= 2 else (-1 if isLeap(year) else -2)
	return epoch - 1 + 365*(year-1) + (year-1)//4 - (year-1)//100 + \
		(year-1)//400 + (367*month-362)//12 + tm + day


def jd_to(jd):
	ordinal = int(jd) - 1721425
	if ordinal > 0:  # > 4x faster
		dt = datetime.fromordinal(ordinal)
		return (dt.year, dt.month, dt.day)

	# wjd = floor(jd - 0.5) + 0.5
	qc, dqc = divmod(jd - epoch, 146097)  # qc ~~ quadricent
	cent, dcent = divmod(dqc, 36524)
	quad, dquad = divmod(dcent, 1461)
	yindex = dquad//365  # divmod(dquad, 365)[0]
	year = qc*400 + cent*100 + quad*4 + yindex + (cent != 4 and yindex != 4)
	yearday = jd - to_jd(year, 1, 1)
	# Python 2.x and 3.x:
	if jd < to_jd(year, 3, 1):
		leapadj = 0
	elif isLeap(year):
		leapadj = 1
	else:
		leapadj = 2
	# Python >= 2.5:
	# leapadj = 0 if jd < to_jd(year, 3, 1) else (1 if isLeap(year) else 2)
	month = ((yearday+leapadj) * 12 + 373) // 367
	day = jd - to_jd(year, month, 1) + 1
	return int(year), int(month), int(day)


def getMonthLen(y, m):
	if m == 12:
		return to_jd(y+1, 1, 1) - to_jd(y, 12, 1)
	else:
		return to_jd(y, m+1, 1) - to_jd(y, m, 1)

J0001 = to_jd(1, 1, 1)
J1970 = to_jd(1970, 1, 1)



def intToBinStr(n, stLen=0):
	bs = b''
	while n > 0:
		bs = bytes([n & 255]) + bs
		n >>= 8
	return bs.rjust(stLen, b'\x00')


def intToBinStr(n, stLen=0):
	return bytes((
		(n >> (i << 3)) & 0xff
		for i in range(int(ceil(log(n, 256)))-1, -1, -1)
	)).rjust(stLen, b'\x00')


def binStrToInt(bs):
	n = 0
	for c in bs:
		n = (n << 8) + c
	return n


def textProgress(n=100, t=0.1):
	import time
	for i in range(n):
		sys.stdout.write('\b\b\b\b##%3d' % (i+1))
		time.sleep(t)
	sys.stdout.write('\b\b\b')


def locate(lst, val):
	n = len(lst)
	if n == 0:
		return
	if val < lst[0]:
		return -0.5
	if val == lst[0]:
		return 0
	if val == lst[-1]:
		return n-1
	if val > lst[-1]:
		return n-0.5
	si = 0  # start index
	ei = n  # end index
	while ei-si > 1:
		mi = (ei+si)/2  # middle index
		if lst[mi] == val:
			return mi
		elif lst[mi] > val:
			ei = mi
			continue
		else:
			si = mi
			continue
	if ei-si == 1:
		return si+0.5


def locate2(lst, val, ind=1):
	n = len(lst)
	if n == 0:
		return
	if val < lst[0][ind]:
		return -0.5
	if val == lst[0][ind]:
		return 0
	if val == lst[-1][ind]:
		return n-1
	if val > lst[-1][ind]:
		return n-0.5
	si = 0
	ei = n
	while ei-si > 1:
		mi = (ei+si)/2
		if lst[mi][ind] == val:
			return mi
		elif lst[mi][ind] > val:
			ei = mi
			continue
		else:
			si = mi
		continue
	if ei-si == 1:
		return si+0.5


def xml2dict(xmlText):
	from xml.etree.ElementTree import XML, tostring
	xmlElems = XML(xmlText)
	for elem in xmlElems:
		elemText = tostring(elem)
		try:
			elem[0]
			elemElems = xml2dict()
		except:
			pass


def sortby(lst, n, reverse=False):
	nlist = [(x[n], x) for x in lst]
	nlist.sort(None, None, reverse)
	return [val for (key, val) in nlist]


def sortby_inplace(lst, n, reverse=False):
	lst[:] = [(x[n], x) for x in lst]
	lst.sort(None, None, reverse)
	lst[:] = [val for (key, val) in lst]
	return


def chBaseIntToStr(number, base):
	"""
		reverse function of int(str, base) and long(str, base)
	"""
	if not 2 <= base <= 36:
		raise ValueError('base must be in 2..36')
	abc = string.digits + string.ascii_letters
	result = ''
	if number < 0:
		number = -number
		sign = '-'
	else:
		sign = ''
	while True:
		number, rdigit = divmod(number, base)
		result = abc[rdigit] + result
		if number == 0:
			return sign + result



#!/usr/bin/env python
# coding: utf-8

"""
	Copyright by https://github.com/zhansliu/writemdict

	pureSalsa20.py -- a pure Python implementation of the Salsa20 cipher, ported to Python 3

	v4.0: Added Python 3 support, dropped support for Python <= 2.5.
	
	// zhansliu

	Original comments below.

	====================================================================
	There are comments here by two authors about three pieces of software:
		comments by Larry Bugbee about
			Salsa20, the stream cipher by Daniel J. Bernstein 
				 (including comments about the speed of the C version) and
			pySalsa20, Bugbee's own Python wrapper for salsa20.c
				 (including some references), and
		comments by Steve Witham about
			pureSalsa20, Witham's pure Python 2.5 implementation of Salsa20,
				which follows pySalsa20's API, and is in this file.

	Salsa20: a Fast Streaming Cipher (comments by Larry Bugbee)
	-----------------------------------------------------------

	Salsa20 is a fast stream cipher written by Daniel Bernstein 
	that basically uses a hash function and XOR making for fast 
	encryption.  (Decryption uses the same function.)  Salsa20 
	is simple and quick.  
	
	Some Salsa20 parameter values...
		design strength    128 bits
		key length         128 or 256 bits, exactly
		IV, aka nonce      64 bits, always
		chunk size         must be in multiples of 64 bytes
	
	Salsa20 has two reduced versions, 8 and 12 rounds each.
	
	One benchmark (10 MB):
		1.5GHz PPC G4     102/97/89 MB/sec for 8/12/20 rounds
		AMD Athlon 2500+   77/67/53 MB/sec for 8/12/20 rounds
		  (no I/O and before Python GC kicks in)
	
	Salsa20 is a Phase 3 finalist in the EU eSTREAM competition 
	and appears to be one of the fastest ciphers.  It is well 
	documented so I will not attempt any injustice here.  Please 
	see "References" below.
	
	...and Salsa20 is "free for any use".  
	
	
	pySalsa20: a Python wrapper for Salsa20 (Comments by Larry Bugbee)
	------------------------------------------------------------------

	pySalsa20.py is a simple ctypes Python wrapper.  Salsa20 is 
	as it's name implies, 20 rounds, but there are two reduced 
	versions, 8 and 12 rounds each.  Because the APIs are 
	identical, pySalsa20 is capable of wrapping all three 
	versions (number of rounds hardcoded), including a special 
	version that allows you to set the number of rounds with a 
	set_rounds() function.  Compile the version of your choice 
	as a shared library (not as a Python extension), name and 
	install it as libsalsa20.so.
	
	Sample usage:
		from pySalsa20 import Salsa20
		s20 = Salsa20(key, IV)
		dataout = s20.encryptBytes(datain)   # same for decrypt
	
	This is EXPERIMENTAL software and intended for educational 
	purposes only.  To make experimentation less cumbersome, 
	pySalsa20 is also free for any use.      
	
	THIS PROGRAM IS PROVIDED WITHOUT WARRANTY OR GUARANTEE OF
	ANY KIND.  USE AT YOUR OWN RISK.  
	
	Enjoy,
	  
	Larry Bugbee
	bugbee@seanet.com
	April 2007

	
	References:
	-----------
	  http://en.wikipedia.org/wiki/Salsa20
	  http://en.wikipedia.org/wiki/Daniel_Bernstein
	  http://cr.yp.to/djb.html
	  http://www.ecrypt.eu.org/stream/salsa20p3.html
	  http://www.ecrypt.eu.org/stream/p3ciphers/salsa20/salsa20_p3source.zip

	 
	Prerequisites for pySalsa20:
	----------------------------
	  - Python 2.5 (haven't tested in 2.4)


	pureSalsa20: Salsa20 in pure Python 2.5 (comments by Steve Witham)
	------------------------------------------------------------------

	pureSalsa20 is the stand-alone Python code in this file.
	It implements the underlying Salsa20 core algorithm
	and emulates pySalsa20's Salsa20 class API (minus a bug(*)).

	pureSalsa20 is MUCH slower than libsalsa20.so wrapped with pySalsa20--
	about 1/1000 the speed for Salsa20/20 and 1/500 the speed for Salsa20/8,
	when encrypting 64k-byte blocks on my computer.

	pureSalsa20 is for cases where portability is much more important than
	speed.  I wrote it for use in a "structured" random number generator.

	There are comments about the reasons for this slowness in
		  http://www.tiac.net/~sw/2010/02/PureSalsa20

	Sample usage:
		from pureSalsa20 import Salsa20
		s20 = Salsa20(key, IV)
		dataout = s20.encryptBytes(datain)   # same for decrypt

	I took the test code from pySalsa20, added a bunch of tests including
	rough speed tests, and moved them into the file testSalsa20.py.  
	To test both pySalsa20 and pureSalsa20, type
		python testSalsa20.py

	(*)The bug (?) in pySalsa20 is this.  The rounds variable is global to the
	libsalsa20.so library and not switched when switching between instances
	of the Salsa20 class.
		s1 = Salsa20( key, IV, 20 )
		s2 = Salsa20( key, IV, 8 )
	In this example,
		with pySalsa20, both s1 and s2 will do 8 rounds of encryption.
		with pureSalsa20, s1 will do 20 rounds and s2 will do 8 rounds.
	Perhaps giving each instance its own nRounds variable, which
	is passed to the salsa20wordtobyte() function, is insecure.  I'm not a 
	cryptographer.

	pureSalsa20.py and testSalsa20.py are EXPERIMENTAL software and 
	intended for educational purposes only.  To make experimentation less 
	cumbersome, pureSalsa20.py and testSalsa20.py are free for any use.

	Revisions:
	----------
	  p3.2   Fixed bug that initialized the output buffer with plaintext!
			 Saner ramping of nreps in speed test.
			 Minor changes and print statements.
	  p3.1   Took timing variability out of add32() and rot32().
			 Made the internals more like pySalsa20/libsalsa .
			 Put the semicolons back in the main loop!
			 In encryptBytes(), modify a byte array instead of appending.
			 Fixed speed calculation bug.
			 Used subclasses instead of patches in testSalsa20.py .
			 Added 64k-byte messages to speed test to be fair to pySalsa20.
	  p3     First version, intended to parallel pySalsa20 version 3.

	More references:
	----------------
	  http://www.seanet.com/~bugbee/crypto/salsa20/          [pySalsa20]
	  http://cr.yp.to/snuffle.html        [The original name of Salsa20]
	  http://cr.yp.to/snuffle/salsafamily-20071225.pdf [ Salsa20 design]
	  http://www.tiac.net/~sw/2010/02/PureSalsa20
	
	THIS PROGRAM IS PROVIDED WITHOUT WARRANTY OR GUARANTEE OF
	ANY KIND.  USE AT YOUR OWN RISK.  

	Cheers,

	Steve Witham sw at remove-this tiac dot net
	February, 2010
"""
import sys
assert(sys.version_info >= (2, 6))

if sys.version_info >= (3,):
	integer_types = (int,)
	python3 = True
else:
	integer_types = (int, long)
	python3 = False

from struct import Struct
little_u64 = Struct( "<Q" )      #	little-endian 64-bit unsigned.
								 #	Unpacks to a tuple of one element!

little16_i32 = Struct( "<16i" )  # 16 little-endian 32-bit signed ints.
little4_i32 = Struct( "<4i" )    #  4 little-endian 32-bit signed ints.
little2_i32 = Struct( "<2i" )    #  2 little-endian 32-bit signed ints.

_version = 'p4.0'

#----------- Salsa20 class which emulates pySalsa20.Salsa20 ---------------

class Salsa20(object):
	def __init__(self, key=None, IV=None, rounds=20 ):
		self._lastChunk64 = True
		self._IVbitlen = 64             # must be 64 bits
		self.ctx = [ 0 ] * 16
		if key:
			self.setKey(key)
		if IV:
			self.setIV(IV)

		self.setRounds(rounds)


	def setKey(self, key):
		assert type(key) == bytes
		ctx = self.ctx
		if len( key ) == 32:  # recommended
			constants = b"expand 32-byte k"
			ctx[ 1],ctx[ 2],ctx[ 3],ctx[ 4] = little4_i32.unpack(key[0:16])
			ctx[11],ctx[12],ctx[13],ctx[14] = little4_i32.unpack(key[16:32])
		elif len( key ) == 16:
			constants = b"expand 16-byte k"
			ctx[ 1],ctx[ 2],ctx[ 3],ctx[ 4] = little4_i32.unpack(key[0:16])
			ctx[11],ctx[12],ctx[13],ctx[14] = little4_i32.unpack(key[0:16])
		else:
			raise Exception( "key length isn't 32 or 16 bytes." )
		ctx[0],ctx[5],ctx[10],ctx[15] = little4_i32.unpack( constants )

		
	def setIV(self, IV):
		assert type(IV) == bytes
		assert len(IV)*8 == 64, 'nonce (IV) not 64 bits'
		self.IV = IV
		ctx=self.ctx
		ctx[ 6],ctx[ 7] = little2_i32.unpack( IV )
		ctx[ 8],ctx[ 9] = 0, 0  # Reset the block counter.

	setNonce = setIV            # support an alternate name


	def setCounter( self, counter ):
		assert( type(counter) in integer_types )
		assert( 0 <= counter < 1<<64 ), "counter < 0 or >= 2**64"
		ctx = self.ctx
		ctx[ 8],ctx[ 9] = little2_i32.unpack( little_u64.pack( counter ) )

	def getCounter( self ):
		return little_u64.unpack( little2_i32.pack( *self.ctx[ 8:10 ] ) ) [0]


	def setRounds(self, rounds, testing=False ):
		assert testing or rounds in [8, 12, 20], 'rounds must be 8, 12, 20'
		self.rounds = rounds


	def encryptBytes(self, data):
		assert type(data) == bytes, 'data must be byte string'
		assert self._lastChunk64, 'previous chunk not multiple of 64 bytes'
		lendata = len(data)
		munged = bytearray(lendata)
		for i in range( 0, lendata, 64 ):
			h = salsa20_wordtobyte( self.ctx, self.rounds, checkRounds=False )
			self.setCounter( ( self.getCounter() + 1 ) % 2**64 )
			# Stopping at 2^70 bytes per nonce is user's responsibility.
			for j in range( min( 64, lendata - i ) ):
				if python3:
					munged[ i+j ] = data[ i+j ] ^ h[j]
				else:
					munged[ i+j ] = ord(data[ i+j ]) ^ ord(h[j])

		self._lastChunk64 = not lendata % 64
		return bytes(munged)
	
	decryptBytes = encryptBytes # encrypt and decrypt use same function

#--------------------------------------------------------------------------

def salsa20_wordtobyte( input, nRounds=20, checkRounds=True ):
	""" Do nRounds Salsa20 rounds on a copy of 
			input: list or tuple of 16 ints treated as little-endian unsigneds.
		Returns a 64-byte string.
		"""

	assert( type(input) in ( list, tuple )  and  len(input) == 16 )
	assert( not(checkRounds) or ( nRounds in [ 8, 12, 20 ] ) )

	x = list( input )

	def XOR( a, b ):  return a ^ b
	ROTATE = rot32
	PLUS   = add32

	for i in range( nRounds // 2 ):
		# These ...XOR...ROTATE...PLUS... lines are from ecrypt-linux.c
		# unchanged except for indents and the blank line between rounds:
		x[ 4] = XOR(x[ 4],ROTATE(PLUS(x[ 0],x[12]), 7));
		x[ 8] = XOR(x[ 8],ROTATE(PLUS(x[ 4],x[ 0]), 9));
		x[12] = XOR(x[12],ROTATE(PLUS(x[ 8],x[ 4]),13));
		x[ 0] = XOR(x[ 0],ROTATE(PLUS(x[12],x[ 8]),18));
		x[ 9] = XOR(x[ 9],ROTATE(PLUS(x[ 5],x[ 1]), 7));
		x[13] = XOR(x[13],ROTATE(PLUS(x[ 9],x[ 5]), 9));
		x[ 1] = XOR(x[ 1],ROTATE(PLUS(x[13],x[ 9]),13));
		x[ 5] = XOR(x[ 5],ROTATE(PLUS(x[ 1],x[13]),18));
		x[14] = XOR(x[14],ROTATE(PLUS(x[10],x[ 6]), 7));
		x[ 2] = XOR(x[ 2],ROTATE(PLUS(x[14],x[10]), 9));
		x[ 6] = XOR(x[ 6],ROTATE(PLUS(x[ 2],x[14]),13));
		x[10] = XOR(x[10],ROTATE(PLUS(x[ 6],x[ 2]),18));
		x[ 3] = XOR(x[ 3],ROTATE(PLUS(x[15],x[11]), 7));
		x[ 7] = XOR(x[ 7],ROTATE(PLUS(x[ 3],x[15]), 9));
		x[11] = XOR(x[11],ROTATE(PLUS(x[ 7],x[ 3]),13));
		x[15] = XOR(x[15],ROTATE(PLUS(x[11],x[ 7]),18));

		x[ 1] = XOR(x[ 1],ROTATE(PLUS(x[ 0],x[ 3]), 7));
		x[ 2] = XOR(x[ 2],ROTATE(PLUS(x[ 1],x[ 0]), 9));
		x[ 3] = XOR(x[ 3],ROTATE(PLUS(x[ 2],x[ 1]),13));
		x[ 0] = XOR(x[ 0],ROTATE(PLUS(x[ 3],x[ 2]),18));
		x[ 6] = XOR(x[ 6],ROTATE(PLUS(x[ 5],x[ 4]), 7));
		x[ 7] = XOR(x[ 7],ROTATE(PLUS(x[ 6],x[ 5]), 9));
		x[ 4] = XOR(x[ 4],ROTATE(PLUS(x[ 7],x[ 6]),13));
		x[ 5] = XOR(x[ 5],ROTATE(PLUS(x[ 4],x[ 7]),18));
		x[11] = XOR(x[11],ROTATE(PLUS(x[10],x[ 9]), 7));
		x[ 8] = XOR(x[ 8],ROTATE(PLUS(x[11],x[10]), 9));
		x[ 9] = XOR(x[ 9],ROTATE(PLUS(x[ 8],x[11]),13));
		x[10] = XOR(x[10],ROTATE(PLUS(x[ 9],x[ 8]),18));
		x[12] = XOR(x[12],ROTATE(PLUS(x[15],x[14]), 7));
		x[13] = XOR(x[13],ROTATE(PLUS(x[12],x[15]), 9));
		x[14] = XOR(x[14],ROTATE(PLUS(x[13],x[12]),13));
		x[15] = XOR(x[15],ROTATE(PLUS(x[14],x[13]),18));

	for i in range( len( input ) ):
		x[i] = PLUS( x[i], input[i] )
	return little16_i32.pack( *x )

#--------------------------- 32-bit ops -------------------------------

def trunc32( w ):
	""" Return the bottom 32 bits of w as a Python int.
		This creates longs temporarily, but returns an int. """
	w = int( ( w & 0x7fffFFFF ) | -( w & 0x80000000 ) )
	assert type(w) == int
	return w


def add32( a, b ):
	""" Add two 32-bit words discarding carry above 32nd bit,
		and without creating a Python long.
		Timing shouldn't vary.
	"""
	lo = ( a & 0xFFFF ) + ( b & 0xFFFF )
	hi = ( a >> 16 ) + ( b >> 16 ) + ( lo >> 16 )
	return ( -(hi & 0x8000) | ( hi & 0x7FFF ) ) << 16 | ( lo & 0xFFFF )


def rot32( w, nLeft ):
	""" Rotate 32-bit word left by nLeft or right by -nLeft
		without creating a Python long.
		Timing depends on nLeft but not on w.
	"""
	nLeft &= 31  # which makes nLeft >= 0
	if nLeft == 0:
		return w

	# Note: now 1 <= nLeft <= 31.
	#	 RRRsLLLLLL   There are nLeft RRR's, (31-nLeft) LLLLLL's,
	# =>  sLLLLLLRRR   and one s which becomes the sign bit.
	RRR = ( ( ( w >> 1 ) & 0x7fffFFFF ) >> ( 31 - nLeft ) )
	sLLLLLL = -( (1<<(31-nLeft)) & w ) | (0x7fffFFFF>>nLeft) & w
	return RRR | ( sLLLLLL << nLeft )


# --------------------------------- end -----------------------------------
